from typing import Optional, Union

from vellum.workflows.constants import AuthorizationType
from vellum.workflows.nodes.displayable.bases.api_node import BaseAPINode
from vellum.workflows.types.core import MergeBehavior, VellumSecret


class APINode(BaseAPINode):
    """
    Used to execute an API call. This node exists to be backwards compatible with Vellum's API Node, and for most cases,
    you should extend from `BaseAPINode` directly.

    url: str - The URL to send the request to.
    method: APIRequestMethod - The HTTP method to use for the request.
    data: Optional[str] - The data to send in the request body.
    json: Optional["JsonObject"] - The JSON data to send in the request body.
    headers: Optional[Dict[str, Union[str, VellumSecret]]] - The headers to send in the request.

    authorization_type: Optional[AuthorizationType] = None - The type of authorization to use for the API call.
    api_key_header_key: Optional[str] = None - The header key to use for the API key authorization.
    bearer_token_value: Optional[Union[str, VellumSecretReference]] = None - The bearer token value to use
    for the bearer token authorization.
    """

    class Display(BaseAPINode.Display):
        icon = "vellum:icon:signal-stream"
        color = "lightBlue"

    authorization_type: Optional[AuthorizationType] = None
    api_key_header_key: Optional[str] = None
    api_key_header_value: Optional[Union[str, VellumSecret]] = None
    bearer_token_value: Optional[Union[str, VellumSecret]] = None

    class Trigger(BaseAPINode.Trigger):
        merge_behavior = MergeBehavior.AWAIT_ANY

    def run(self) -> BaseAPINode.Outputs:
        self._validate()

        headers = self.headers or {}
        header_overrides = {}
        bearer_token = None

        if (
            self.authorization_type == AuthorizationType.API_KEY
            and self.api_key_header_key
            and self.api_key_header_value
        ):
            header_overrides[self.api_key_header_key] = self.api_key_header_value
        elif self.authorization_type == AuthorizationType.BEARER_TOKEN and isinstance(self.bearer_token_value, str):
            header_overrides["Authorization"] = f"Bearer {self.bearer_token_value}"
        elif self.authorization_type == AuthorizationType.BEARER_TOKEN and isinstance(
            self.bearer_token_value, VellumSecret
        ):
            bearer_token = self.bearer_token_value

        final_headers = {**headers, **header_overrides}

        vellum_client_wrapper = self._context.vellum_client._client_wrapper
        if self.url.startswith(vellum_client_wrapper._environment.default) and (
            "X-API-Key" not in final_headers and "X_API_KEY" not in final_headers
        ):
            final_headers["X-API-Key"] = vellum_client_wrapper.api_key

        return self._run(
            method=self.method,
            url=self.url,
            data=self.data,
            json=self.json,
            headers=final_headers,
            bearer_token=bearer_token,
            timeout=self.timeout,
        )



from enum import Enum
from multiprocessing.synchronize import Event as MultiprocessingEvent
from threading import Event as ThreadingEvent
from typing import (  # type: ignore[attr-defined]
    Any,
    Dict,
    List,
    Union,
    _GenericAlias,
    _SpecialGenericAlias,
    _UnionGenericAlias,
)

from vellum.client.core.pydantic_utilities import UniversalBaseModel

JsonArray = List["Json"]
JsonObject = Dict[str, "Json"]
Json = Union[None, bool, int, float, str, JsonArray, JsonObject]

CancelSignal = Union[ThreadingEvent, MultiprocessingEvent]

# Unions and Generics inherit from `_GenericAlias` instead of `type`
# In future versions of python, we'll see `_UnionGenericAlias`
UnderGenericAlias = _GenericAlias
SpecialGenericAlias = _SpecialGenericAlias
UnionGenericAlias = _UnionGenericAlias


class VellumSecret(UniversalBaseModel):
    name: str


EntityInputsInterface = Dict[str, Any]


class MergeBehavior(Enum):
    AWAIT_ALL = "AWAIT_ALL"
    AWAIT_ANY = "AWAIT_ANY"
    AWAIT_ATTRIBUTES = "AWAIT_ATTRIBUTES"
    CUSTOM = "CUSTOM"


class ConditionType(Enum):
    IF = "IF"
    ELIF = "ELIF"
    ELSE = "ELSE"



import dataclasses
import inspect
from typing import TYPE_CHECKING, Annotated, Any, Callable, List, Literal, Optional, Type, Union, get_args, get_origin

from pydantic import BaseModel
from pydantic_core import PydanticUndefined
from pydash import snake_case

from vellum import Vellum
from vellum.client.types.function_definition import FunctionDefinition
from vellum.workflows.integrations.composio_service import ComposioService
from vellum.workflows.integrations.mcp_service import MCPService
from vellum.workflows.integrations.vellum_integration_service import VellumIntegrationService
from vellum.workflows.types.definition import (
    ComposioToolDefinition,
    DeploymentDefinition,
    MCPServer,
    MCPToolDefinition,
    VellumIntegrationToolDefinition,
)
from vellum.workflows.utils.vellum_variables import vellum_variable_type_to_openapi_type

if TYPE_CHECKING:
    from vellum.workflows.workflows.base import BaseWorkflow

type_map: dict[Any, str] = {
    str: "string",
    int: "integer",
    float: "number",
    bool: "boolean",
    list: "array",
    dict: "object",
    None: "null",
    type(None): "null",
    inspect._empty: "null",
    "None": "null",
}

for k, v in list(type_map.items()):
    if isinstance(k, type):
        type_map[k.__name__] = v


def compile_annotation(annotation: Optional[Any], defs: dict[str, Any]) -> dict:
    if annotation is None:
        return {"type": "null"}

    if get_origin(annotation) is Union:
        return {"anyOf": [compile_annotation(a, defs) for a in get_args(annotation)]}

    if get_origin(annotation) is Literal:
        values = list(get_args(annotation))
        types = {type(value) for value in values}
        if len(types) == 1:
            value_type = types.pop()
            if value_type in type_map:
                return {"type": type_map[value_type], "enum": values}
            else:
                return {"enum": values}
        else:
            return {"enum": values}

    if get_origin(annotation) is dict:
        _, value_type = get_args(annotation)
        return {"type": "object", "additionalProperties": compile_annotation(value_type, defs)}

    if get_origin(annotation) is list:
        item_type = get_args(annotation)[0]
        return {"type": "array", "items": compile_annotation(item_type, defs)}

    if get_origin(annotation) is tuple:
        args = get_args(annotation)
        if len(args) == 2 and args[1] is Ellipsis:
            # Tuple[int, ...] with homogeneous items
            return {"type": "array", "items": compile_annotation(args[0], defs)}
        else:
            # Tuple[int, str] with fixed length items
            result = {
                "type": "array",
                "prefixItems": [compile_annotation(arg, defs) for arg in args],
                "minItems": len(args),
                "maxItems": len(args),
            }
            return result

    if dataclasses.is_dataclass(annotation) and isinstance(annotation, type):
        if annotation.__name__ not in defs:
            properties = {}
            required = []
            for field in dataclasses.fields(annotation):
                properties[field.name] = compile_annotation(field.type, defs)
                if field.default is dataclasses.MISSING:
                    required.append(field.name)
                else:
                    properties[field.name]["default"] = _compile_default_value(field.default)
            defs[annotation.__name__] = {"type": "object", "properties": properties, "required": required}
        return {"$ref": f"#/$defs/{annotation.__name__}"}

    if inspect.isclass(annotation) and issubclass(annotation, BaseModel):
        if annotation.__name__ not in defs:
            properties = {}
            required = []
            for field_name, field_info in annotation.model_fields.items():
                # field_info is a FieldInfo object which has an annotation attribute
                properties[field_name] = compile_annotation(field_info.annotation, defs)

                if field_info.description is not None:
                    properties[field_name]["description"] = field_info.description

                if field_info.default is PydanticUndefined:
                    required.append(field_name)
                else:
                    properties[field_name]["default"] = _compile_default_value(field_info.default)
            defs[annotation.__name__] = {"type": "object", "properties": properties, "required": required}

        return {"$ref": f"#/$defs/{annotation.__name__}"}

    return {"type": type_map[annotation]}


def _compile_default_value(default: Any) -> Any:
    if dataclasses.is_dataclass(default):
        return {
            field.name: _compile_default_value(getattr(default, field.name)) for field in dataclasses.fields(default)
        }

    if isinstance(default, BaseModel):
        return {
            field_name: _compile_default_value(getattr(default, field_name))
            for field_name in default.__class__.model_fields.keys()
        }

    return default


def _compile_workflow_deployment_input(input_var: Any) -> dict[str, Any]:
    """
    Converts a deployment workflow input variable to a JSON schema type definition.
    """
    primitive_type = vellum_variable_type_to_openapi_type(input_var.type)
    input_schema = {"type": primitive_type}

    if input_var.default is not None:
        input_schema["default"] = input_var.default.value

    return input_schema


def compile_function_definition(function: Callable) -> FunctionDefinition:
    """
    Converts a Python function into our Vellum-native FunctionDefinition type.

    Args:
        function: The Python function to compile
    """

    try:
        signature = inspect.signature(function)
    except ValueError as e:
        raise ValueError(f"Failed to get signature for function {function.__name__}: {str(e)}")

    # Get inputs from the decorator if present
    inputs = getattr(function, "__vellum_inputs__", {})
    exclude_params = set(inputs.keys())

    properties = {}
    required = []
    defs: dict[str, Any] = {}
    for param in signature.parameters.values():
        # Skip parameters that are in the exclude_params set
        if exclude_params and param.name in exclude_params:
            continue

        # Check if parameter uses Annotated type hint
        if get_origin(param.annotation) is Annotated:
            args = get_args(param.annotation)
            actual_type = args[0]
            # Extract description from metadata
            description = args[1] if len(args) > 1 and isinstance(args[1], str) else None

            properties[param.name] = compile_annotation(actual_type, defs)
            if description:
                properties[param.name]["description"] = description
        else:
            properties[param.name] = compile_annotation(param.annotation, defs)

        if param.default is inspect.Parameter.empty:
            required.append(param.name)
        else:
            properties[param.name]["default"] = _compile_default_value(param.default)

    parameters = {"type": "object", "properties": properties, "required": required}
    if defs:
        parameters["$defs"] = defs

    return FunctionDefinition(
        name=function.__name__,
        description=function.__doc__,
        parameters=parameters,
    )


def compile_inline_workflow_function_definition(workflow_class: Type["BaseWorkflow"]) -> FunctionDefinition:
    """
    Converts a base workflow class into our Vellum-native FunctionDefinition type.
    """
    # Get the inputs class for the workflow
    inputs_class = workflow_class.get_inputs_class()
    vars_inputs_class = vars(inputs_class)

    properties = {}
    required = []
    defs: dict[str, Any] = {}

    for name, field_type in inputs_class.__annotations__.items():
        if name.startswith("__"):
            continue

        properties[name] = compile_annotation(field_type, defs)

        # Check if the field has a default value
        if name not in vars_inputs_class:
            required.append(name)
        else:
            # Field has a default value
            properties[name]["default"] = vars_inputs_class[name]

    parameters = {"type": "object", "properties": properties, "required": required}
    if defs:
        parameters["$defs"] = defs

    return FunctionDefinition(
        name=snake_case(workflow_class.__name__),
        description=workflow_class.__doc__,
        parameters=parameters,
    )


def compile_workflow_deployment_function_definition(
    deployment_definition: DeploymentDefinition,
    vellum_client: Vellum,
) -> FunctionDefinition:
    """
    Converts a deployment workflow config into our Vellum-native FunctionDefinition type.

    Args:
        deployment_definition: DeploymentDefinition instance
        vellum_client: Vellum client instance
    """
    release_info = deployment_definition.get_release_info(vellum_client)

    name = release_info["name"]
    description = release_info["description"]
    input_variables = release_info["input_variables"]

    properties = {}
    required = []

    for input_var in input_variables:
        properties[input_var.key] = _compile_workflow_deployment_input(input_var)

        if input_var.required and input_var.default is None:
            required.append(input_var.key)

    parameters = {"type": "object", "properties": properties, "required": required}

    return FunctionDefinition(
        name=name.replace("-", ""),
        description=description,
        parameters=parameters,
    )


def get_mcp_tool_name(tool_def: MCPToolDefinition) -> str:
    """Generate a unique name for an MCP tool by combining server and tool names."""
    server_name = snake_case(tool_def.server.name)
    return f"{server_name}__{tool_def.name}"


def compile_mcp_tool_definition(server_def: MCPServer) -> List[MCPToolDefinition]:
    """Hydrate an MCPToolDefinition with detailed information from the MCP server.

    We do tool discovery on the MCP server to get the tool definitions.

    Args:
        server_def: The basic MCPToolDefinition to enhance

    Returns:
        MCPToolDefinition with detailed parameters and description
    """
    try:
        mcp_service = MCPService()
        return mcp_service.hydrate_tool_definitions(server_def)
    except Exception:
        return []


def compile_composio_tool_definition(tool_def: ComposioToolDefinition) -> FunctionDefinition:
    """Hydrate a ComposioToolDefinition with detailed information from the Composio API.

    Args:
        tool_def: The basic ComposioToolDefinition to enhance

    Returns:
        FunctionDefinition with detailed parameters and description
    """
    try:
        composio_service = ComposioService()
        tool_details = composio_service.get_tool_by_slug(tool_def.action)

        # Create a FunctionDefinition directly with proper field extraction
        return FunctionDefinition(
            name=tool_def.name,
            description=tool_details.get("description", tool_def.description),
            parameters=tool_details.get("input_parameters", {}),
        )
    except Exception:
        # If hydration fails (including no API key), return basic function definition
        return FunctionDefinition(
            name=tool_def.name,
            description=tool_def.description,
            parameters={},
        )


def compile_vellum_integration_tool_definition(
    tool_def: VellumIntegrationToolDefinition,
    vellum_client: Vellum,
) -> FunctionDefinition:
    """Compile a VellumIntegrationToolDefinition into a FunctionDefinition.

    Args:
        tool_def: The VellumIntegrationToolDefinition to compile
        vellum_client: Vellum client instance

    Returns:
        FunctionDefinition with tool parameters and description
    """
    try:
        service = VellumIntegrationService(vellum_client)
        tool_details = service.get_tool_definition(
            integration=tool_def.integration_name,
            provider=tool_def.provider.value,
            tool_name=tool_def.name,
        )

        return FunctionDefinition(
            name=tool_def.name,
            description=tool_details.description,
            parameters=tool_details.parameters or {},
        )
    except Exception:
        # Fallback for service failures
        return FunctionDefinition(name=tool_def.name, description=tool_def.description, parameters={})


def use_tool_inputs(**inputs):
    """
    Decorator to specify which parameters of a tool function should be provided
    from the parent workflow inputs rather than from the LLM.

    Args:
        **inputs: Mapping of parameter names to parent input references

    Example:
        @use_tool_inputs(
            parent_input=ParentInputs.parent_input,
        )
        def get_string(parent_input: str, user_query: str) -> str:
            return f"Parent: {parent_input}, Query: {user_query}"
    """

    def decorator(func: Callable) -> Callable:
        # Store the inputs mapping on the function for later use
        setattr(func, "__vellum_inputs__", inputs)
        return func

    return decorator



import json
import logging
import os
import re
import subprocess
from typing import List, Optional

import docker
from docker import DockerClient
from dotenv import load_dotenv

from vellum.client.core.api_error import ApiError
from vellum.workflows.vellum_client import create_vellum_client, create_vellum_environment
from vellum_cli.config import DEFAULT_WORKSPACE_CONFIG, load_vellum_cli_config
from vellum_cli.logger import handle_cli_error, load_cli_logger

_SUPPORTED_ARCHITECTURE = "amd64"


def image_push_command(
    image: str, tags: Optional[List[str]] = None, workspace: Optional[str] = None, source: Optional[str] = None
) -> None:
    load_dotenv(dotenv_path=os.path.join(os.getcwd(), ".env"))
    logger = load_cli_logger()
    config = load_vellum_cli_config()

    logger.info("Cleaning up unused Docker images...")
    try:
        subprocess.run(
            ["docker", "image", "prune", "--all", "-f", "--filter", "label=image-type=python-workflow-runtime"],
            check=True,
        )
        logger.info("Docker image pruning completed successfully")
    except subprocess.CalledProcessError as e:
        logger.warning(f"Docker image pruning failed: {e}")

    if source:
        logger.info(f"Building Docker image from Dockerfile: {source}")

        if not os.path.exists(source):
            handle_cli_error(logger, "Dockerfile not found", f"Dockerfile does not exist: {source}")

        build_result = subprocess.run(
            ["docker", "buildx", "build", "-f", source, "--platform=linux/amd64", "-t", image, "."],
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
        )

        if build_result.returncode != 0:
            handle_cli_error(logger, "Docker build failed", build_result.stderr.decode("utf-8"))

        logger.info("Docker build completed successfully")
    workspace_config = next((w for w in config.workspaces if w.name == workspace), DEFAULT_WORKSPACE_CONFIG)

    api_key = os.getenv(workspace_config.api_key, None)
    if not api_key:
        raise ValueError(f"API key {workspace_config.api_key} for workspace {workspace} not found")

    vellum_client = create_vellum_client(
        api_key=api_key,
        api_url=workspace_config.api_url,
    )

    # Check if we are self hosted by looking at our base url
    api_url = create_vellum_environment().default
    is_self_hosted = not api_url.endswith("vellum.ai") and "localhost:" not in api_url and "127.0.0.1" not in api_url
    if is_self_hosted:
        logger.info("Self hosted install detected...")

    if is_self_hosted and "/" not in image:
        logger.error(
            "For adding images to your self hosted install you must include the "
            "repository address in the provided image name. You must also have "
            "already pushed the image to the docker repository that your self "
            "hosted install is using."
        )
        exit(1)

    # We're using docker python SDK here instead of subprocess since it connects to the docker host directly
    # instead of using the command line so it seemed like it would possibly be a little more robust since
    # it might avoid peoples' wonky paths, unfortunately it doesn't support the manifest command which we need for
    # listing all of the architectures of the image instead of just the one that matches the machine. We can fall back
    # to using normal inspect which returns the machine image for this case though. And in the future we could figure
    # out how to call the docker host directly to do this.
    logger.info("Pre-validating image...")
    docker_client = docker.from_env()
    check_architecture(docker_client, image, logger)

    repo_split = image.split("/")
    tag_split = repo_split[-1].split(":")
    image_name = tag_split[0]
    main_tag = tag_split[1] if len(tag_split) > 1 else "latest"
    all_tags = [main_tag, *(tags or [])]

    if is_self_hosted:
        # Include the repo for self hosted since we skip pushing it to our internal repo and expect
        # the user the push it themselves and provide us with the repo name.
        image_name = image.split(":")[0]

        logger.info("Checking if image already exists in repository...")
        exists_result = subprocess.run(
            ["docker", "manifest", "inspect", image],
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
        )

        if exists_result.returncode != 0:
            logger.error(
                "Image does not exist in repository. Push the image to the "
                "repository your self hosted install is using and try again."
            )
            exit(1)
    else:
        logger.info("Authenticating...")
        try:
            auth = vellum_client.container_images.docker_service_token()
        except ApiError as e:
            if e.status_code == 401 or e.status_code == 403:
                handle_cli_error(
                    logger,
                    title="Authentication failed",
                    message="Unable to authenticate with Vellum API. Please check your API key.",
                    suggestion="Make sure your VELLUM_API_KEY environment variable is set correctly.",
                )
                return
            elif e.status_code == 400 and isinstance(e.body, dict) and "detail" in e.body:
                handle_cli_error(
                    logger,
                    title="API request failed",
                    message=e.body["detail"],
                )
                return
            elif e.status_code == 500:
                handle_cli_error(
                    logger,
                    title="Server error",
                    message="The Vellum API failed with an unexpected server error.",
                    suggestion="Please try again later and contact support if the problem persists.",
                )
                return
            else:
                handle_cli_error(
                    logger,
                    title="API request failed",
                    message=f"Failed to get Docker service token from Vellum API (HTTP {e.status_code}).",
                    suggestion="Please check your configuration and try again. If the problem persists, "
                    "contact support.",
                )
                return

        docker_client.login(
            username="oauth2accesstoken",
            password=auth.access_token,
            registry=auth.repository,
        )

        for tag in all_tags:
            vellum_image_name = f"{auth.repository}/{image_name}:{tag}"

            docker_client.api.tag(image, vellum_image_name)

            push_result = docker_client.images.push(repository=vellum_image_name, stream=True)

            # Here were trying to mime the output you would get from a normal docker push, which
            # the python sdk makes as hard as possible.
            for raw_line in push_result:
                try:
                    for sub_line in raw_line.decode("utf-8").split("\r\n"):
                        line = json.loads(sub_line)
                        error_message = line.get("errorDetail", {}).get("message")
                        status = line.get("status")
                        id = line.get("id", "")

                        if error_message:
                            logger.error(error_message)
                            exit(1)
                        elif status == "Waiting":
                            continue
                        elif status:
                            logger.info(f"{id}{': ' if id else ''}{status}")
                        else:
                            logger.info(line)
                except Exception:
                    continue

    inspect_result = subprocess.run(
        ["docker", "inspect", "--format='{{index .RepoDigests 0}}'", image],
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
    )

    sha = ""
    if inspect_result.returncode == 0:
        match = re.search(r"sha256[^']*", inspect_result.stdout.decode("utf-8"))
        if match and match.group(0):
            sha = match.group(0)

    if not sha:
        # Fallback to using docker client if command line fails, at least on some systems
        # this appears to give a bad sha.
        logger.warning(
            "Could not determine image hash with command line docker falling back to python docker client..."
        )

        image_details = docker_client.api.inspect_image(image)
        sha = image_details["Id"]

    logger.info(f"Updating Vellum metadata and validating image works in our system with image digest: {sha}...")

    try:
        vellum_client.container_images.push_container_image(
            name=image_name,
            sha=sha,
            tags=all_tags,
        )
    except ApiError as e:
        if e.status_code == 401 or e.status_code == 403:
            handle_cli_error(
                logger,
                title="Authentication failed",
                message="Unable to push container image metadata to Vellum API. Please check your API key.",
                suggestion="Make sure your VELLUM_API_KEY environment variable is set correctly.",
            )
            return
        elif e.status_code == 400 and isinstance(e.body, dict) and "detail" in e.body:
            handle_cli_error(
                logger,
                title="API request failed",
                message=e.body["detail"],
            )
            return
        elif e.status_code == 500:
            handle_cli_error(
                logger,
                title="Server error",
                message="The Vellum API failed with an unexpected server error while pushing container image metadata.",
                suggestion="Please try again later and contact support if the problem persists.",
            )
            return
        else:
            handle_cli_error(
                logger,
                title="API request failed",
                message=f"Failed to push container image metadata to Vellum API (HTTP {e.status_code}).",
                suggestion="Please check your configuration and try again. If the problem persists, contact support.",
            )
            return

    logger.info(f"Image successfully pushed as {image_name} to vellum with tags: {all_tags}.")


def check_architecture(docker_client: DockerClient, image: str, logger: logging.Logger) -> None:
    result = subprocess.run(
        ["docker", "manifest", "inspect", image],
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
    )

    manifest_parse_failed = False
    architectures = []
    try:
        manifest = json.loads(result.stdout)
        architectures = [manifest_item["platform"]["architecture"] for manifest_item in manifest["manifests"]]
    except Exception:
        logger.debug("Error parsing manifest response")
        manifest_parse_failed = True

    # Fall back to inspect image if we errored out using docker command line
    if result.returncode != 0 or manifest_parse_failed:
        logger.debug(f"Error inspecting manifest: {result.stderr.decode('utf-8').strip()}")
        image_details = docker_client.api.inspect_image(image)

        if image_details["Architecture"] != _SUPPORTED_ARCHITECTURE:
            logger.error(f"Image must be built for {_SUPPORTED_ARCHITECTURE} architecture.")
            exit(1)
    else:
        if _SUPPORTED_ARCHITECTURE not in architectures:
            logger.error(f"Image must be built for {_SUPPORTED_ARCHITECTURE} architecture.")
            exit(1)



from typing import Any, Dict, Iterator, Set, Tuple, Type, Union, get_args, get_origin
from typing_extensions import dataclass_transform

from pydantic import GetCoreSchemaHandler
from pydantic.fields import FieldInfo
from pydantic_core import core_schema

from vellum.workflows.constants import undefined
from vellum.workflows.errors.types import WorkflowErrorCode
from vellum.workflows.exceptions import WorkflowInitializationException
from vellum.workflows.references import ExternalInputReference, WorkflowInputReference
from vellum.workflows.references.input import InputReference
from vellum.workflows.types.utils import get_class_attr_names, infer_types


@dataclass_transform(kw_only_default=True)
class _BaseInputsMeta(type):
    def __new__(cls, name: str, bases: Tuple[Type, ...], dct: Dict[str, Any]) -> Any:
        dct["__parent_class__"] = type(None)
        return super().__new__(cls, name, bases, dct)

    def __getattribute__(cls, name: str) -> Any:
        if name.startswith("_") or not issubclass(cls, BaseInputs):
            return super().__getattribute__(name)

        attr_names = get_class_attr_names(cls)
        if name in attr_names:
            # We first try to resolve the instance that this class attribute name is mapped to. If it's not found,
            # we iterate through its inheritance hierarchy to find the first base class that has this attribute
            # and use its mapping.
            instance = vars(cls).get(name, undefined)
            if instance is undefined:
                for base in cls.__mro__[1:]:
                    inherited_input_reference = getattr(base, name, undefined)
                    if isinstance(inherited_input_reference, (ExternalInputReference, WorkflowInputReference)):
                        instance = inherited_input_reference.instance
                        break

            types = infer_types(cls, name)
            if getattr(cls, "__descriptor_class__", None) is ExternalInputReference:
                return ExternalInputReference(name=name, types=types, instance=instance, inputs_class=cls)
            else:
                return WorkflowInputReference(name=name, types=types, instance=instance, inputs_class=cls)

        if getattr(cls, "__descriptor_class__", None) is ExternalInputReference:
            return ExternalInputReference(name=name, types=(), instance=undefined, inputs_class=cls)
        else:
            return WorkflowInputReference(name=name, types=(), instance=undefined, inputs_class=cls)

    def __iter__(cls) -> Iterator[InputReference]:
        # We iterate through the inheritance hierarchy to find all the WorkflowInputReference attached to this
        # Inputs class. __mro__ is the method resolution order, which is the order in which base classes are resolved.
        yielded_attr_names: Set[str] = set()

        for resolved_cls in cls.__mro__:
            attr_names = get_class_attr_names(resolved_cls)
            for attr_name in attr_names:
                if attr_name in yielded_attr_names:
                    continue

                attr_value = getattr(resolved_cls, attr_name)
                if not isinstance(attr_value, (WorkflowInputReference, ExternalInputReference)):
                    continue

                yield attr_value
                yielded_attr_names.add(attr_name)


class BaseInputs(metaclass=_BaseInputsMeta):
    __parent_class__: Type = type(None)

    def __init__(self, **kwargs: Any) -> None:
        """
        Initialize BaseInputs with provided keyword arguments.

        Validation logic:
        1. Ensures all required fields (values without defaults, regardless of Optional or not) either:
        - Have a value provided in kwargs, or
        - Have a default value defined in the class
        2. Validates that no None values are provided for required fields
        3. Sets all provided values as attributes on the instance

        Args:
            **kwargs: Keyword arguments corresponding to the class's type annotations

        Raises:
            WorkflowInitializationException: If a required field is missing or None
        """
        for name, field_type in self.__class__.__annotations__.items():
            # Get the value (either from kwargs or class default)
            value = kwargs.get(name, undefined)
            has_default = name in vars(self.__class__)

            # If no value provided, check for default
            if value is undefined and has_default:
                default_value = vars(self.__class__)[name]
                # Check if default is a FieldInfo with default_factory
                if isinstance(default_value, FieldInfo):
                    if default_value.default_factory is not None:
                        # Call the factory function to create a new instance
                        value = default_value.default_factory()  # type: ignore[call-arg]
                    elif hasattr(default_value, "default") and default_value.default is not ...:
                        # Use the default value directly
                        value = default_value.default
                else:
                    # Regular default value
                    value = default_value

            # Check if field_type allows None
            origin = get_origin(field_type)
            args = get_args(field_type)
            is_optional = origin is Union and type(None) in args

            if value is undefined and not has_default:
                # All fields without defaults must be provided, even if Optional
                raise WorkflowInitializationException(
                    message=f"Required input variables {name} should have defined value",
                    code=WorkflowErrorCode.INVALID_INPUTS,
                    workflow_definition=self.__class__.__parent_class__,
                )

            # Validate that None is not provided for non-Optional fields
            if value is None and not is_optional:
                raise WorkflowInitializationException(
                    message=f"Required input variables {name} should have defined value",
                    code=WorkflowErrorCode.INVALID_INPUTS,
                    workflow_definition=self.__class__.__parent_class__,
                )

            # Set the value on the instance (either from kwargs or default)
            if value is not undefined:
                setattr(self, name, value)

    def __iter__(self) -> Iterator[Tuple[InputReference, Any]]:
        for input_descriptor in self.__class__:
            if hasattr(self, input_descriptor.name):
                yield (input_descriptor, getattr(self, input_descriptor.name))

    def __vellum_encode__(self) -> Dict[str, Any]:
        """Return a JSON-serializable representation of this inputs object."""
        return {descriptor.name: value for descriptor, value in self if value is not undefined}

    @classmethod
    def __get_pydantic_core_schema__(
        cls, source_type: Type[Any], handler: GetCoreSchemaHandler
    ) -> core_schema.CoreSchema:
        return core_schema.is_instance_schema(cls)



from vellum_ee.workflows.display.workflows.get_vellum_workflow_display_class import get_workflow_display

from tests.workflows.basic_search_node.workflow import BasicSearchWorkflow


def test_serialize_workflow():
    # GIVEN a Workflow with a search node
    # WHEN we serialize it

    workflow_display = get_workflow_display(workflow_class=BasicSearchWorkflow)

    serialized_workflow: dict = workflow_display.serialize()

    # THEN we should get a serialized representation of the workflow
    assert serialized_workflow.keys() == {
        "workflow_raw_data",
        "input_variables",
        "state_variables",
        "output_variables",
    }

    # AND its input variables should be what we expect
    input_variables = serialized_workflow["input_variables"]
    assert len(input_variables) == 1
    assert input_variables == [
        {
            "id": "6e405c6c-36eb-4c06-9d54-ae06cccce585",
            "key": "query",
            "type": "STRING",
            "default": None,
            "required": True,
            "extensions": {"color": None},
        }
    ]

    # AND its output variables should be what we expect
    output_variables = serialized_workflow["output_variables"]
    assert len(output_variables) == 1
    assert output_variables == [{"id": "27424f7d-9767-4059-bdcf-c2be8b798fd7", "key": "text", "type": "STRING"}]

    # AND its raw data is what we expect
    workflow_raw_data = serialized_workflow["workflow_raw_data"]
    assert len(workflow_raw_data["nodes"]) == 3
    assert len(workflow_raw_data["edges"]) == 2

    # AND each node should be serialized correctly
    entrypoint_node = workflow_raw_data["nodes"][0]
    assert entrypoint_node == {
        "id": "06671b25-5c6b-4675-8c74-6c396a608728",
        "type": "ENTRYPOINT",
        "inputs": [],
        "data": {"label": "Entrypoint Node", "source_handle_id": "df80b4aa-2ba1-49a2-8375-fb1f78eee31f"},
        "display_data": {"position": {"x": 0.0, "y": -50.0}},
        "base": None,
        "definition": None,
    }

    search_node = workflow_raw_data["nodes"][1]
    assert search_node == {
        "id": "89c8bee0-8015-4d73-9112-e436ab086567",
        "type": "SEARCH",
        "inputs": [
            {
                "id": "1dbfd1d9-8e8e-47b6-bb65-c578385ef978",
                "key": "query",
                "value": {
                    "rules": [
                        {
                            "type": "INPUT_VARIABLE",
                            "data": {"input_variable_id": "6e405c6c-36eb-4c06-9d54-ae06cccce585"},
                        }
                    ],
                    "combinator": "OR",
                },
            },
            {
                "id": "8f197069-4729-4db3-811c-9a522685dfba",
                "key": "document_index_id",
                "value": {
                    "rules": [{"type": "CONSTANT_VALUE", "data": {"type": "STRING", "value": "name"}}],
                    "combinator": "OR",
                },
            },
            {
                "id": "dd991fc6-5690-496b-b63e-b1a1219f3682",
                "key": "weights",
                "value": {
                    "rules": [
                        {
                            "type": "CONSTANT_VALUE",
                            "data": {"type": "JSON", "value": {"keywords": 0.45, "semantic_similarity": 0.55}},
                        }
                    ],
                    "combinator": "OR",
                },
            },
            {
                "id": "52f6c3ab-cd96-4ca7-bfb1-7554c431b318",
                "key": "limit",
                "value": {
                    "rules": [{"type": "CONSTANT_VALUE", "data": {"type": "JSON", "value": None}}],
                    "combinator": "OR",
                },
            },
            {
                "id": "e1822f91-ca9c-46c9-8dcd-cce83fe331ce",
                "key": "separator",
                "value": {
                    "rules": [{"type": "CONSTANT_VALUE", "data": {"type": "STRING", "value": "\n\n#####\n\n"}}],
                    "combinator": "OR",
                },
            },
            {
                "id": "3ee2bc2a-d72a-471a-8dd6-5e5028db8bfe",
                "key": "result_merging_enabled",
                "value": {
                    "rules": [{"type": "CONSTANT_VALUE", "data": {"type": "STRING", "value": "False"}}],
                    "combinator": "OR",
                },
            },
            {
                "id": "5c766934-13f7-4a0f-a751-21e25fd2ca30",
                "key": "external_id_filters",
                "value": {
                    "rules": [{"type": "CONSTANT_VALUE", "data": {"type": "JSON", "value": None}}],
                    "combinator": "OR",
                },
            },
            {
                "id": "977d75bc-520f-447b-ad75-8b56c793f951",
                "key": "metadata_filters",
                "value": {
                    "rules": [
                        {
                            "type": "CONSTANT_VALUE",
                            "data": {
                                "type": "JSON",
                                "value": {
                                    "type": "LOGICAL_CONDITION_GROUP",
                                    "combinator": "AND",
                                    "conditions": [
                                        {
                                            "type": "LOGICAL_CONDITION",
                                            "lhs_variable_id": "a6322ca2-8b65-4d26-b3a1-f926dcada0fa",
                                            "operator": "=",
                                            "rhs_variable_id": "c539a2e2-0873-43b0-ae21-81790bb1c4cb",
                                        },
                                        {
                                            "type": "LOGICAL_CONDITION",
                                            "lhs_variable_id": "a89483b6-6850-4105-8c4e-ec0fd197cd43",
                                            "operator": "=",
                                            "rhs_variable_id": "847b8ee0-2c37-4e41-9dea-b4ba3579e2c1",
                                        },
                                    ],
                                    "negated": False,
                                },
                            },
                        }
                    ],
                    "combinator": "OR",
                },
            },
            {
                "id": "a6322ca2-8b65-4d26-b3a1-f926dcada0fa",
                "key": "vellum-query-builder-variable-a6322ca2-8b65-4d26-b3a1-f926dcada0fa",
                "value": {
                    "rules": [
                        {
                            "type": "INPUT_VARIABLE",
                            "data": {"input_variable_id": "67fd852b-789b-4c18-8465-2bbf1696b8eb"},
                        }
                    ],
                    "combinator": "OR",
                },
            },
            {
                "id": "c539a2e2-0873-43b0-ae21-81790bb1c4cb",
                "key": "vellum-query-builder-variable-c539a2e2-0873-43b0-ae21-81790bb1c4cb",
                "value": {
                    "rules": [
                        {
                            "type": "INPUT_VARIABLE",
                            "data": {"input_variable_id": "58e909b0-9269-454d-b40d-3846fd2c39f2"},
                        }
                    ],
                    "combinator": "OR",
                },
            },
            {
                "id": "a89483b6-6850-4105-8c4e-ec0fd197cd43",
                "key": "vellum-query-builder-variable-a89483b6-6850-4105-8c4e-ec0fd197cd43",
                "value": {
                    "rules": [
                        {
                            "type": "INPUT_VARIABLE",
                            "data": {"input_variable_id": "7564f9c7-b2cf-4584-b4c7-845a14ac4dfa"},
                        }
                    ],
                    "combinator": "OR",
                },
            },
            {
                "id": "847b8ee0-2c37-4e41-9dea-b4ba3579e2c1",
                "key": "vellum-query-builder-variable-847b8ee0-2c37-4e41-9dea-b4ba3579e2c1",
                "value": {
                    "rules": [
                        {
                            "type": "INPUT_VARIABLE",
                            "data": {"input_variable_id": "cd081063-ab26-4ec7-99b3-af7e035a16e3"},
                        }
                    ],
                    "combinator": "OR",
                },
            },
        ],
        "data": {
            "label": "Simple Search Node",
            "results_output_id": "184c5214-29be-4029-8ece-2991972e0822",
            "text_output_id": "9bab7f1b-3a4b-46bf-8b30-e3016ac38f51",
            "error_output_id": None,
            "source_handle_id": "928c79f2-bc07-43ee-9420-380a3bd36fb8",
            "target_handle_id": "85db938d-9a85-4f08-8bbf-b795db2c40d5",
            "query_node_input_id": "1dbfd1d9-8e8e-47b6-bb65-c578385ef978",
            "document_index_node_input_id": "8f197069-4729-4db3-811c-9a522685dfba",
            "weights_node_input_id": "dd991fc6-5690-496b-b63e-b1a1219f3682",
            "limit_node_input_id": "52f6c3ab-cd96-4ca7-bfb1-7554c431b318",
            "separator_node_input_id": "e1822f91-ca9c-46c9-8dcd-cce83fe331ce",
            "result_merging_enabled_node_input_id": "3ee2bc2a-d72a-471a-8dd6-5e5028db8bfe",
            "external_id_filters_node_input_id": "5c766934-13f7-4a0f-a751-21e25fd2ca30",
            "metadata_filters_node_input_id": "977d75bc-520f-447b-ad75-8b56c793f951",
        },
        "display_data": {"position": {"x": 200.0, "y": -50.0}},
        "base": {
            "name": "SearchNode",
            "module": ["vellum", "workflows", "nodes", "displayable", "search_node", "node"],
        },
        "definition": {
            "name": "SimpleSearchNode",
            "module": ["tests", "workflows", "basic_search_node", "workflow"],
        },
        "trigger": {
            "id": "85db938d-9a85-4f08-8bbf-b795db2c40d5",
            "merge_behavior": "AWAIT_ANY",
        },
        "ports": [{"id": "928c79f2-bc07-43ee-9420-380a3bd36fb8", "name": "default", "type": "DEFAULT"}],
    }

    final_output_node = workflow_raw_data["nodes"][2]
    assert final_output_node == {
        "id": "4e466510-6756-403f-a182-56e5a2b85d94",
        "type": "TERMINAL",
        "data": {
            "label": "Final Output",
            "name": "text",
            "target_handle_id": "cd8c736f-1b77-493d-b857-d8feb5c03b15",
            "output_id": "27424f7d-9767-4059-bdcf-c2be8b798fd7",
            "output_type": "STRING",
            "node_input_id": "39c7f674-a794-4525-8a04-f22a40ed0914",
        },
        "inputs": [
            {
                "id": "39c7f674-a794-4525-8a04-f22a40ed0914",
                "key": "node_input",
                "value": {
                    "rules": [
                        {
                            "type": "NODE_OUTPUT",
                            "data": {
                                "node_id": "89c8bee0-8015-4d73-9112-e436ab086567",
                                "output_id": "9bab7f1b-3a4b-46bf-8b30-e3016ac38f51",
                            },
                        }
                    ],
                    "combinator": "OR",
                },
            }
        ],
        "display_data": {"position": {"x": 400.0, "y": -50.0}},
        "base": {
            "name": "FinalOutputNode",
            "module": ["vellum", "workflows", "nodes", "displayable", "final_output_node", "node"],
        },
        "definition": None,
    }

    # AND each edge should be serialized correctly
    serialized_edges = workflow_raw_data["edges"]
    assert serialized_edges == [
        {
            "id": "16500e66-1e19-4ee3-a197-baf9c2926ae4",
            "source_node_id": "06671b25-5c6b-4675-8c74-6c396a608728",
            "source_handle_id": "df80b4aa-2ba1-49a2-8375-fb1f78eee31f",
            "target_node_id": "89c8bee0-8015-4d73-9112-e436ab086567",
            "target_handle_id": "85db938d-9a85-4f08-8bbf-b795db2c40d5",
            "type": "DEFAULT",
        },
        {
            "id": "cb918deb-f546-47b5-8b6b-db0d22a29fd1",
            "source_node_id": "89c8bee0-8015-4d73-9112-e436ab086567",
            "source_handle_id": "928c79f2-bc07-43ee-9420-380a3bd36fb8",
            "target_node_id": "4e466510-6756-403f-a182-56e5a2b85d94",
            "target_handle_id": "cd8c736f-1b77-493d-b857-d8feb5c03b15",
            "type": "DEFAULT",
        },
    ]

    # AND the display data is what we expect
    display_data = workflow_raw_data["display_data"]
    assert display_data == {"viewport": {"x": 0.0, "y": 0.0, "zoom": 1.0}}

    # AND the definition is what we expect
    definition = workflow_raw_data["definition"]
    assert definition == {
        "name": "BasicSearchWorkflow",
        "module": [
            "tests",
            "workflows",
            "basic_search_node",
            "workflow",
        ],
    }



from dataclasses import asdict, is_dataclass
import inspect
from uuid import UUID
from typing import TYPE_CHECKING, Any, Dict, List, Optional, cast

from pydantic import BaseModel

from vellum.client.types.logical_operator import LogicalOperator
from vellum.workflows.constants import undefined
from vellum.workflows.descriptors.base import BaseDescriptor
from vellum.workflows.expressions.accessor import AccessorExpression
from vellum.workflows.expressions.add import AddExpression
from vellum.workflows.expressions.and_ import AndExpression
from vellum.workflows.expressions.begins_with import BeginsWithExpression
from vellum.workflows.expressions.between import BetweenExpression
from vellum.workflows.expressions.coalesce_expression import CoalesceExpression
from vellum.workflows.expressions.concat import ConcatExpression
from vellum.workflows.expressions.contains import ContainsExpression
from vellum.workflows.expressions.does_not_begin_with import DoesNotBeginWithExpression
from vellum.workflows.expressions.does_not_contain import DoesNotContainExpression
from vellum.workflows.expressions.does_not_end_with import DoesNotEndWithExpression
from vellum.workflows.expressions.does_not_equal import DoesNotEqualExpression
from vellum.workflows.expressions.ends_with import EndsWithExpression
from vellum.workflows.expressions.equals import EqualsExpression
from vellum.workflows.expressions.greater_than import GreaterThanExpression
from vellum.workflows.expressions.greater_than_or_equal_to import GreaterThanOrEqualToExpression
from vellum.workflows.expressions.in_ import InExpression
from vellum.workflows.expressions.is_error import IsErrorExpression
from vellum.workflows.expressions.is_nil import IsNilExpression
from vellum.workflows.expressions.is_not_nil import IsNotNilExpression
from vellum.workflows.expressions.is_not_null import IsNotNullExpression
from vellum.workflows.expressions.is_not_undefined import IsNotUndefinedExpression
from vellum.workflows.expressions.is_null import IsNullExpression
from vellum.workflows.expressions.is_undefined import IsUndefinedExpression
from vellum.workflows.expressions.length import LengthExpression
from vellum.workflows.expressions.less_than import LessThanExpression
from vellum.workflows.expressions.less_than_or_equal_to import LessThanOrEqualToExpression
from vellum.workflows.expressions.minus import MinusExpression
from vellum.workflows.expressions.not_between import NotBetweenExpression
from vellum.workflows.expressions.not_in import NotInExpression
from vellum.workflows.expressions.or_ import OrExpression
from vellum.workflows.expressions.parse_json import ParseJsonExpression
from vellum.workflows.nodes.bases.base import BaseNode
from vellum.workflows.nodes.displayable.bases.utils import primitive_to_vellum_value
from vellum.workflows.nodes.utils import get_unadorned_node
from vellum.workflows.references.constant import ConstantValueReference
from vellum.workflows.references.environment_variable import EnvironmentVariableReference
from vellum.workflows.references.execution_count import ExecutionCountReference
from vellum.workflows.references.lazy import LazyReference
from vellum.workflows.references.output import OutputReference
from vellum.workflows.references.state_value import StateValueReference
from vellum.workflows.references.trigger import TriggerAttributeReference
from vellum.workflows.references.vellum_secret import VellumSecretReference
from vellum.workflows.references.workflow_input import WorkflowInputReference
from vellum.workflows.types.core import JsonArray, JsonObject
from vellum.workflows.types.generics import is_workflow_class
from vellum.workflows.utils.files import virtual_open
from vellum.workflows.utils.functions import compile_function_definition
from vellum.workflows.utils.uuids import uuid4_from_hash
from vellum_ee.workflows.display.utils.exceptions import InvalidInputReferenceError, UnsupportedSerializationException

if TYPE_CHECKING:
    from vellum_ee.workflows.display.types import WorkflowDisplayContext


def convert_descriptor_to_operator(descriptor: BaseDescriptor) -> LogicalOperator:
    if isinstance(descriptor, EqualsExpression):
        return "="
    elif isinstance(descriptor, DoesNotEqualExpression):
        return "!="
    elif isinstance(descriptor, LessThanExpression):
        return "<"
    elif isinstance(descriptor, GreaterThanExpression):
        return ">"
    elif isinstance(descriptor, LessThanOrEqualToExpression):
        return "<="
    elif isinstance(descriptor, GreaterThanOrEqualToExpression):
        return ">="
    elif isinstance(descriptor, ContainsExpression):
        return "contains"
    elif isinstance(descriptor, BeginsWithExpression):
        return "beginsWith"
    elif isinstance(descriptor, EndsWithExpression):
        return "endsWith"
    elif isinstance(descriptor, DoesNotContainExpression):
        return "doesNotContain"
    elif isinstance(descriptor, DoesNotBeginWithExpression):
        return "doesNotBeginWith"
    elif isinstance(descriptor, DoesNotEndWithExpression):
        return "doesNotEndWith"
    elif isinstance(descriptor, IsErrorExpression):
        return "isError"
    elif isinstance(descriptor, (IsNullExpression, IsNilExpression, IsUndefinedExpression)):
        return "null"
    elif isinstance(descriptor, (IsNotNullExpression, IsNotNilExpression, IsNotUndefinedExpression)):
        return "notNull"
    elif isinstance(descriptor, InExpression):
        return "in"
    elif isinstance(descriptor, NotInExpression):
        return "notIn"
    elif isinstance(descriptor, BetweenExpression):
        return "between"
    elif isinstance(descriptor, NotBetweenExpression):
        return "notBetween"
    elif isinstance(descriptor, AndExpression):
        return "and"
    elif isinstance(descriptor, OrExpression):
        return "or"
    elif isinstance(descriptor, CoalesceExpression):
        return "coalesce"
    elif isinstance(descriptor, ParseJsonExpression):
        return "parseJson"
    elif isinstance(descriptor, LengthExpression):
        return "length"
    elif isinstance(descriptor, AddExpression):
        return "+"
    elif isinstance(descriptor, MinusExpression):
        return "-"
    elif isinstance(descriptor, ConcatExpression):
        return "concat"
    else:
        raise ValueError(f"Unsupported descriptor type: {descriptor}")


def get_child_descriptor(value: LazyReference, display_context: "WorkflowDisplayContext") -> BaseDescriptor:
    if isinstance(value._get, str):
        reference_parts = value._get.split(".")
        if len(reference_parts) < 3:
            raise Exception(f"Failed to parse lazy reference: {value._get}. Only Node Output references are supported.")

        output_name = reference_parts[-1]
        nested_class_name = reference_parts[-2]
        if nested_class_name != "Outputs":
            raise Exception(
                f"Failed to parse lazy reference: {value._get}. Outputs are the only node reference supported."
            )

        node_class_name = ".".join(reference_parts[:-2])
        for node in display_context.global_node_displays.keys():
            if node.__name__ == node_class_name:
                return getattr(node.Outputs, output_name)

        raise Exception(f"Failed to parse lazy reference: {value._get}")

    return value._get()


def _get_pydantic_model_definition(model_class: type) -> Optional[JsonObject]:
    """Extract module path definition from a Pydantic model class."""
    if not inspect.isclass(model_class) or not issubclass(model_class, BaseModel):
        return None

    module = model_class.__module__
    name = model_class.__name__

    module_path = module.split(".")

    return {
        "name": name,
        "module": cast(JsonArray, module_path),
    }


def _serialize_condition(
    executable_id: UUID, display_context: "WorkflowDisplayContext", condition: BaseDescriptor
) -> JsonObject:
    if isinstance(
        condition,
        (
            IsErrorExpression,
            IsNullExpression,
            IsNotNullExpression,
            IsNilExpression,
            IsNotNilExpression,
            IsUndefinedExpression,
            IsNotUndefinedExpression,
            LengthExpression,
            ParseJsonExpression,
        ),
    ):
        lhs = serialize_value(executable_id, display_context, condition._expression)
        return {
            "type": "UNARY_EXPRESSION",
            "lhs": lhs,
            "operator": convert_descriptor_to_operator(condition),
        }
    elif isinstance(condition, (BetweenExpression, NotBetweenExpression)):
        base = serialize_value(executable_id, display_context, condition._value)
        lhs = serialize_value(executable_id, display_context, condition._start)
        rhs = serialize_value(executable_id, display_context, condition._end)

        return {
            "type": "TERNARY_EXPRESSION",
            "base": base,
            "operator": convert_descriptor_to_operator(condition),
            "lhs": lhs,
            "rhs": rhs,
        }
    elif isinstance(
        condition,
        (
            AddExpression,
            AndExpression,
            BeginsWithExpression,
            CoalesceExpression,
            ConcatExpression,
            ContainsExpression,
            DoesNotBeginWithExpression,
            DoesNotContainExpression,
            DoesNotEndWithExpression,
            DoesNotEqualExpression,
            EndsWithExpression,
            EqualsExpression,
            GreaterThanExpression,
            GreaterThanOrEqualToExpression,
            InExpression,
            LessThanExpression,
            LessThanOrEqualToExpression,
            MinusExpression,
            NotInExpression,
            OrExpression,
        ),
    ):
        lhs = serialize_value(executable_id, display_context, condition._lhs)
        rhs = serialize_value(executable_id, display_context, condition._rhs)

        return {
            "type": "BINARY_EXPRESSION",
            "lhs": lhs,
            "operator": convert_descriptor_to_operator(condition),
            "rhs": rhs,
        }
    elif isinstance(condition, AccessorExpression):
        return {
            "type": "BINARY_EXPRESSION",
            "lhs": serialize_value(executable_id, display_context, condition._base),
            "operator": "accessField",
            "rhs": serialize_value(executable_id, display_context, condition._field),
        }

    raise UnsupportedSerializationException(f"Unsupported condition type: {condition.__class__.__name__}")


def serialize_key(key: Any) -> str:
    """Serialize dictionary keys to strings, handling function objects properly."""
    if callable(key):
        return key.__name__
    elif isinstance(key, str):
        return key
    else:
        return str(key)


def serialize_value(executable_id: UUID, display_context: "WorkflowDisplayContext", value: Any) -> Optional[JsonObject]:
    """
    Serialize a value to a JSON object. Returns `None` if the value resolves to `undefined`.
    This is safe because all valid values are a JSON object, including the `None` constant:
    > `{"type": "CONSTANT_VALUE", "value": {"type": "JSON", "value": None}}`

    Args:
        executable_id: node id or workflow id
        display_context: workflow display context
        value: value to serialize

    Returns:
        serialized value
    """
    if value is undefined:
        return None

    if isinstance(value, ConstantValueReference):
        return serialize_value(executable_id, display_context, value._value)

    if isinstance(value, LazyReference):
        child_descriptor = get_child_descriptor(value, display_context)
        return serialize_value(executable_id, display_context, child_descriptor)

    if isinstance(value, WorkflowInputReference):
        if value not in display_context.global_workflow_input_displays:
            workflow_inputs_class = display_context.workflow_display_class.infer_workflow_class().get_inputs_class()
            if value.inputs_class != workflow_inputs_class:
                inputs_class_name = value.inputs_class.__name__
                workflow_class_name = display_context.workflow_display_class.infer_workflow_class().__name__
                raise UnsupportedSerializationException(
                    f"Inputs class '{inputs_class_name}' referenced during serialization of "
                    f"'{workflow_class_name}' without parameterizing this Workflow with this Inputs definition. "
                    f"Update your Workflow definition to "
                    f"'{workflow_class_name}(BaseWorkflow[{inputs_class_name}, BaseState])'."
                )
            else:
                raise InvalidInputReferenceError(
                    message=f"Inputs class '{value.inputs_class.__qualname__}' has no attribute '{value.name}'",
                    inputs_class_name=value.inputs_class.__qualname__,
                    attribute_name=value.name,
                )
        workflow_input_display = display_context.global_workflow_input_displays[value]
        return {
            "type": "WORKFLOW_INPUT",
            "input_variable_id": str(workflow_input_display.id),
        }

    if isinstance(value, StateValueReference):
        state_value_display = display_context.global_state_value_displays[value]
        return {
            "type": "WORKFLOW_STATE",
            "state_variable_id": str(state_value_display.id),
        }

    if isinstance(value, OutputReference):
        output_display = display_context.global_node_output_displays[value]

        upstream_node_class = value.outputs_class.__parent_class__
        if not issubclass(upstream_node_class, BaseNode):
            raise ValueError(f"Output references must be to a node, not {upstream_node_class}")
        unadorned_upstream_node_class = get_unadorned_node(upstream_node_class)
        upstream_node = display_context.global_node_displays[unadorned_upstream_node_class]

        return {
            "type": "NODE_OUTPUT",
            "node_id": str(upstream_node.node_id),
            "node_output_id": str(output_display.id),
        }

    if isinstance(value, VellumSecretReference):
        return {
            "type": "VELLUM_SECRET",
            "vellum_secret_name": value.name,
        }

    if isinstance(value, EnvironmentVariableReference):
        if value.serialize_as_constant:
            return {
                "type": "CONSTANT_VALUE",
                "value": {
                    "type": "STRING",
                    "value": value.name,
                },
            }
        else:
            return {
                "type": "ENVIRONMENT_VARIABLE",
                "environment_variable": value.name,
            }

    if isinstance(value, ExecutionCountReference):
        node_class_display = display_context.global_node_displays[value.node_class]

        return {
            "type": "EXECUTION_COUNTER",
            "node_id": str(node_class_display.node_id),
        }

    if isinstance(value, TriggerAttributeReference):
        trigger_class = value.trigger_class
        trigger_id = trigger_class.__id__

        return {
            "type": "TRIGGER_ATTRIBUTE",
            "trigger_id": str(trigger_id),
            "attribute_id": str(value.id),
        }

    if isinstance(value, list):
        serialized_items = []
        for item in value:
            serialized_item = serialize_value(executable_id, display_context, item)
            if serialized_item is not None:
                serialized_items.append(serialized_item)

        if all(isinstance(item, dict) and item["type"] == "CONSTANT_VALUE" for item in serialized_items):
            constant_values = []
            for item in serialized_items:
                item_dict = cast(Dict[str, Any], item)
                value_inner = item_dict["value"]

                if value_inner["type"] == "JSON" and "items" in value_inner:
                    # Nested JSON list
                    constant_values.append(value_inner["items"])
                else:
                    constant_values.append(value_inner["value"])

            return {
                "type": "CONSTANT_VALUE",
                "value": {
                    "type": "JSON",
                    "value": constant_values,
                },
            }
        else:
            return {
                "type": "ARRAY_REFERENCE",
                "items": cast(JsonArray, serialized_items),  # list[JsonObject] -> JsonArray
            }

    if is_dataclass(value) and not isinstance(value, type):
        dict_value = asdict(value)
        return serialize_value(executable_id, display_context, dict_value)

    if isinstance(value, dict):
        serialized_entries: List[Dict[str, Any]] = []
        for key, val in value.items():
            serialized_val = serialize_value(executable_id, display_context, val)
            if serialized_val is not None:
                serialized_entries.append(
                    {
                        "id": str(uuid4_from_hash(f"{executable_id}|{key}")),
                        "key": serialize_key(key),
                        "value": serialized_val,
                    }
                )

        # Check if all entries have constant values
        if all(entry["value"]["type"] == "CONSTANT_VALUE" for entry in serialized_entries):
            constant_entries = {}
            for entry in serialized_entries:
                entry_value = entry["value"]["value"]
                constant_entries[entry["key"]] = entry_value["value"]

            return {
                "type": "CONSTANT_VALUE",
                "value": {
                    "type": "JSON",
                    "value": constant_entries,
                },
            }
        else:
            return {
                "type": "DICTIONARY_REFERENCE",
                "entries": cast(JsonArray, serialized_entries),
            }

    if is_workflow_class(value):
        from vellum_ee.workflows.display.workflows.get_vellum_workflow_display_class import get_workflow_display

        # Pass the parent display context so the subworkflow can resolve parent workflow inputs
        workflow_display = get_workflow_display(
            workflow_class=value,
            parent_display_context=display_context,
        )
        serialized_value: dict = workflow_display.serialize()
        name = serialized_value["workflow_raw_data"]["definition"]["name"]
        description = value.__doc__ or ""
        return {
            "type": "CONSTANT_VALUE",
            "value": {
                "type": "JSON",
                "value": {
                    "type": "INLINE_WORKFLOW",
                    "name": name,
                    "description": description,
                    "exec_config": serialized_value,
                },
            },
        }

    if isinstance(value, BaseModel):
        context = {"executable_id": executable_id, "client": display_context.client}
        dict_value = value.model_dump(context=context)
        dict_ref = serialize_value(executable_id, display_context, dict_value)
        if dict_ref is not None and dict_ref.get("type") == "DICTIONARY_REFERENCE":
            dict_ref["definition"] = _get_pydantic_model_definition(value.__class__)
        return dict_ref

    if callable(value):
        function_definition = compile_function_definition(value)

        name = function_definition.name
        description = function_definition.description or ""

        inputs = getattr(value, "__vellum_inputs__", {})

        if inputs:
            serialized_inputs = {}
            for param_name, input_ref in inputs.items():
                serialized_input = serialize_value(executable_id, display_context, input_ref)
                if serialized_input is not None:
                    serialized_inputs[param_name] = serialized_input

            model_data = function_definition.model_dump()
            model_data["inputs"] = serialized_inputs
            function_definition_data = model_data
        else:
            function_definition_data = function_definition.model_dump()

        source_path = inspect.getsourcefile(value)
        if source_path is not None:
            with virtual_open(source_path) as f:
                source_code = f.read()
        else:
            source_code = f"Source code not available for {value.__name__}"

        return {
            "type": "CONSTANT_VALUE",
            "value": {
                "type": "JSON",
                "value": {
                    "type": "CODE_EXECUTION",
                    "name": name,
                    "description": description,
                    "definition": function_definition_data,
                    "src": source_code,
                },
            },
        }

    if not isinstance(value, BaseDescriptor):
        vellum_value = primitive_to_vellum_value(value)
        return {
            "type": "CONSTANT_VALUE",
            "value": vellum_value.dict(),
        }

    # If it's not any of the references we know about,
    # then try to serialize it as a nested value
    return _serialize_condition(executable_id, display_context, value)



# This file was auto-generated by Fern from our API Definition.

import datetime as dt
import typing

import pydantic
from ..core.pydantic_utilities import IS_PYDANTIC_V2, UniversalBaseModel
from .document_processing_state import DocumentProcessingState
from .document_status import DocumentStatus
from .processing_failure_reason_enum import ProcessingFailureReasonEnum
from .slim_document_document_to_document_index import SlimDocumentDocumentToDocumentIndex


class SlimDocument(UniversalBaseModel):
    id: str = pydantic.Field()
    """
    Vellum-generated ID that uniquely identifies this document.
    """

    external_id: typing.Optional[str] = pydantic.Field(default=None)
    """
    The external ID that was originally provided when uploading the document.
    """

    last_uploaded_at: dt.datetime = pydantic.Field()
    """
    A timestamp representing when this document was most recently uploaded.
    """

    label: str = pydantic.Field()
    """
    Human-friendly name for this document.
    """

    processing_state: DocumentProcessingState
    processing_failure_reason: typing.Optional[ProcessingFailureReasonEnum] = pydantic.Field(default=None)
    """
    An enum value representing why the document could not be processed. Is null unless processing_state is FAILED.
    
    * `EXCEEDED_CHARACTER_LIMIT` - Exceeded Character Limit
    * `INVALID_FILE` - Invalid File
    * `INVALID_CREDENTIALS` - Invalid Credentials
    """

    status: typing.Optional[DocumentStatus] = pydantic.Field(default=None)
    """
    The document's current status.
    
    * `ACTIVE` - Active
    """

    keywords: typing.Optional[typing.List[str]] = pydantic.Field(default=None)
    """
    A list of keywords associated with this document. Originally provided when uploading the document.
    """

    metadata: typing.Optional[typing.Dict[str, typing.Optional[typing.Any]]] = pydantic.Field(default=None)
    """
    A previously supplied JSON object containing metadata that can be filtered on when searching.
    """

    document_to_document_indexes: typing.List[SlimDocumentDocumentToDocumentIndex]

    if IS_PYDANTIC_V2:
        model_config: typing.ClassVar[pydantic.ConfigDict] = pydantic.ConfigDict(extra="allow", frozen=True)  # type: ignore # Pydantic v2
    else:

        class Config:
            frozen = True
            smart_union = True
            extra = pydantic.Extra.allow



"""Descriptor for referring to trigger attributes in workflow graphs."""

from __future__ import annotations

from uuid import UUID
from typing import TYPE_CHECKING, Any, Generic, Optional, Tuple, Type, TypeVar, cast

from pydantic import GetCoreSchemaHandler
from pydantic_core import core_schema

from vellum.workflows.descriptors.base import BaseDescriptor
from vellum.workflows.errors.types import WorkflowErrorCode
from vellum.workflows.exceptions import NodeException
from vellum.workflows.utils.uuids import get_trigger_attribute_id

if TYPE_CHECKING:
    from vellum.workflows.state.base import BaseState
    from vellum.workflows.triggers.base import BaseTrigger

_T = TypeVar("_T")


class TriggerAttributeReference(BaseDescriptor[_T], Generic[_T]):
    """Reference to a trigger attribute defined via type annotations."""

    def __init__(
        self,
        *,
        name: str,
        types: Tuple[Type[_T], ...],
        instance: Optional[_T],
        trigger_class: Type[BaseTrigger],
    ) -> None:
        super().__init__(name=name, types=types, instance=instance)
        self._trigger_class = trigger_class

    @property
    def trigger_class(self) -> Type[BaseTrigger]:
        return self._trigger_class

    @property
    def id(self) -> UUID:
        """Generate deterministic UUID from trigger class qualname and attribute name."""
        return get_trigger_attribute_id(self._trigger_class, self.name)

    def resolve(self, state: BaseState) -> _T:
        trigger_attributes = getattr(state.meta, "trigger_attributes", {})
        if self in trigger_attributes:
            return cast(_T, trigger_attributes[self])

        if state.meta.parent:
            return self.resolve(state.meta.parent)

        if type(None) in self.types:
            return cast(_T, None)

        raise NodeException(
            message=f"Missing trigger attribute '{self.name}' for {self._trigger_class.__name__}",
            code=WorkflowErrorCode.INVALID_INPUTS,
        )

    def __repr__(self) -> str:
        return f"{self._trigger_class.__qualname__}.{self.name}"

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, TriggerAttributeReference):
            return False
        return super().__eq__(other) and self._trigger_class == other._trigger_class

    def __hash__(self) -> int:
        return hash((self._trigger_class, self._name))

    @classmethod
    def __get_pydantic_core_schema__(
        cls, source_type: Type[Any], handler: GetCoreSchemaHandler
    ) -> core_schema.CoreSchema:
        return core_schema.is_instance_schema(cls)



import inspect
import os
from typing import Any, ClassVar, Dict, Generic, List, Optional, Sequence, Tuple, Type, TypeVar, cast, get_args

from vellum import (
    ArrayInput,
    ChatHistoryInput,
    ChatMessage,
    CodeExecutionPackage,
    CodeExecutionRuntime,
    CodeExecutorInput,
    ErrorInput,
    FunctionCall,
    FunctionCallInput,
    JsonInput,
    NumberInput,
    SearchResult,
    SearchResultsInput,
    StringInput,
    VellumError,
    VellumValue,
)
from vellum.client.core import RequestOptions
from vellum.client.core.api_error import ApiError
from vellum.client.types.code_executor_secret_input import CodeExecutorSecretInput
from vellum.workflows.constants import undefined
from vellum.workflows.errors.types import WorkflowErrorCode
from vellum.workflows.exceptions import NodeException
from vellum.workflows.nodes.bases import BaseNode
from vellum.workflows.nodes.bases.base import BaseNodeMeta
from vellum.workflows.nodes.displayable.bases.utils import primitive_to_vellum_value
from vellum.workflows.nodes.displayable.code_execution_node.utils import read_file_from_path, run_code_inline
from vellum.workflows.outputs.base import BaseOutputs
from vellum.workflows.types.core import EntityInputsInterface, MergeBehavior, VellumSecret
from vellum.workflows.types.generics import StateType
from vellum.workflows.types.utils import get_original_base
from vellum.workflows.utils.vellum_variables import primitive_type_to_vellum_variable_type

_OutputType = TypeVar("_OutputType")


# TODO: Consolidate all dynamic output metaclasses
# https://app.shortcut.com/vellum/story/5533
class _CodeExecutionNodeMeta(BaseNodeMeta):
    def __new__(mcs, name: str, bases: Tuple[Type, ...], dct: Dict[str, Any]) -> Any:
        parent = super().__new__(mcs, name, bases, dct)

        # We use the compiled class to infer the output type for the Outputs.result descriptor.
        if not isinstance(parent, _CodeExecutionNodeMeta):
            raise ValueError("CodeExecutionNode must be created with the CodeExecutionNodeMeta metaclass")

        annotations = parent.__dict__["Outputs"].__annotations__
        parent.__dict__["Outputs"].__annotations__ = {
            **annotations,
            "result": parent.get_output_type(),
        }
        return parent

    def get_output_type(cls) -> Type:
        original_base = get_original_base(cls)
        all_args = get_args(original_base)

        if len(all_args) < 2 or isinstance(all_args[1], TypeVar):
            return str
        else:
            return all_args[1]


class CodeExecutionNode(BaseNode[StateType], Generic[StateType, _OutputType], metaclass=_CodeExecutionNodeMeta):
    """
    Used to execute an arbitrary script. This node exists to be backwards compatible with
    Vellum's Code Execution Node, and for most cases, you should extend from `BaseNode` directly.

    filepath: str - The path to the script to execute.
    code_inputs: EntityInputsInterface - The inputs for the custom script.
    runtime: CodeExecutionRuntime = "PYTHON_3_12" - The runtime to use for the custom script.
    packages: Optional[Sequence[CodeExecutionPackage]] = None - The packages to use for the custom script.
    request_options: Optional[RequestOptions] = None - The request options to use for the custom script.
    """

    class Display(BaseNode.Display):
        icon = "vellum:icon:rectangle-code"
        color = "lime"

    filepath: ClassVar[Optional[str]] = None
    code: ClassVar[Optional[str]] = None

    code_inputs: ClassVar[EntityInputsInterface] = {}
    runtime: CodeExecutionRuntime = "PYTHON_3_11_6"
    packages: Optional[Sequence[CodeExecutionPackage]] = None

    request_options: Optional[RequestOptions] = None

    class Trigger(BaseNode.Trigger):
        merge_behavior = MergeBehavior.AWAIT_ANY

    class Outputs(BaseOutputs):
        # We use our mypy plugin to override the _OutputType with the actual output type
        # for downstream references to this output.
        result: _OutputType  # type: ignore[valid-type]
        log: str

    def run(self) -> Outputs:
        output_type = self.__class__.get_output_type()
        code, filepath = self._resolve_code()
        if not self.packages and self.runtime == "PYTHON_3_11_6" and not self._has_secrets_in_code_inputs():
            logs, result = run_code_inline(code, self.code_inputs, output_type, filepath)
            return self.Outputs(result=result, log=logs)

        else:
            input_values = self._compile_code_inputs()
            expected_output_type = primitive_type_to_vellum_variable_type(output_type)

            try:
                code_execution_result = self._context.vellum_client.execute_code(
                    input_values=input_values,
                    code=code,
                    runtime=self.runtime,
                    output_type=expected_output_type,
                    packages=self.packages or [],
                    request_options=self.request_options,
                )
            except ApiError as e:
                self._handle_api_error(e)

            if code_execution_result.output.type != expected_output_type:
                actual_type = code_execution_result.output.type
                raise NodeException(
                    code=WorkflowErrorCode.INVALID_OUTPUTS,
                    message=f"Expected an output of type '{expected_output_type}', received '{actual_type}'",
                )

            return self.Outputs(result=code_execution_result.output.value, log=code_execution_result.log)

    def _handle_api_error(self, e: ApiError) -> None:
        if e.status_code and e.status_code == 403 and isinstance(e.body, dict):
            raise NodeException(
                message=e.body.get("detail", "Provider credentials is missing or unavailable"),
                code=WorkflowErrorCode.PROVIDER_CREDENTIALS_UNAVAILABLE,
            ) from e

        if e.status_code and e.status_code >= 400 and e.status_code < 500 and isinstance(e.body, dict):
            raise NodeException(
                message=e.body.get("message", e.body.get("detail", "Failed to execute code")),
                code=WorkflowErrorCode.INVALID_INPUTS,
            ) from e

        raise NodeException(
            message="Failed to execute code",
            code=WorkflowErrorCode.INTERNAL_ERROR,
        ) from e

    def _has_secrets_in_code_inputs(self) -> bool:
        """Check if any code_inputs contain VellumSecret instances that require API execution."""
        for input_value in self.code_inputs.values():
            if isinstance(input_value, VellumSecret):
                return True
        return False

    def _compile_code_inputs(self) -> List[CodeExecutorInput]:
        # TODO: We may want to consolidate with prompt deployment input compilation
        # https://app.shortcut.com/vellum/story/4117

        compiled_inputs: List[CodeExecutorInput] = []

        for input_name, input_value in self.code_inputs.items():
            if input_value is undefined:
                continue
            if isinstance(input_value, str):
                compiled_inputs.append(
                    StringInput(
                        name=input_name,
                        value=input_value,
                    )
                )
            elif isinstance(input_value, VellumSecret):
                compiled_inputs.append(
                    CodeExecutorSecretInput(
                        name=input_name,
                        value=input_value.name,
                    )
                )
            elif isinstance(input_value, list):
                if all(isinstance(message, ChatMessage) for message in input_value):
                    compiled_inputs.append(
                        ChatHistoryInput(
                            name=input_name,
                            value=cast(List[ChatMessage], input_value),
                        )
                    )
                elif all(isinstance(message, SearchResult) for message in input_value):
                    compiled_inputs.append(
                        SearchResultsInput(
                            name=input_name,
                            value=cast(List[SearchResult], input_value),
                        )
                    )
                else:
                    # Convert primitive values to VellumValue objects
                    vellum_values: List[VellumValue] = [primitive_to_vellum_value(item) for item in input_value]

                    compiled_inputs.append(
                        ArrayInput(
                            name=input_name,
                            value=vellum_values,
                        )
                    )
            elif isinstance(input_value, dict):
                compiled_inputs.append(
                    JsonInput(
                        name=input_name,
                        value=cast(Dict[str, Any], input_value),
                    )
                )
            elif isinstance(input_value, (float, int)):
                compiled_inputs.append(
                    NumberInput(
                        name=input_name,
                        value=float(input_value),
                    )
                )
            elif isinstance(input_value, FunctionCall):
                compiled_inputs.append(
                    FunctionCallInput(
                        name=input_name,
                        value=cast(FunctionCall, input_value),
                    )
                )
            elif isinstance(input_value, VellumError):
                compiled_inputs.append(
                    ErrorInput(
                        name=input_name,
                        value=cast(VellumError, input_value),
                    )
                )
            else:
                raise NodeException(
                    message=f"Unrecognized input type for input '{input_name}': {input_value.__class__.__name__}",
                    code=WorkflowErrorCode.INVALID_INPUTS,
                )

        return compiled_inputs

    def _resolve_code(self) -> Tuple[str, str]:
        if self.code and self.filepath:
            raise NodeException(
                message="Cannot specify both `code` and `filepath` for a CodeExecutionNode",
                code=WorkflowErrorCode.INVALID_INPUTS,
            )

        if self.code:
            return self.code, f"{self.__class__.__name__}.code.py"

        if not self.filepath:
            raise NodeException(
                message="Must specify either `code` or `filepath` for a CodeExecutionNode",
                code=WorkflowErrorCode.INVALID_INPUTS,
            )

        root = inspect.getfile(self.__class__)

        code = read_file_from_path(node_filepath=root, script_filepath=self.filepath, context=self._context)
        if not code:
            raise NodeException(
                message=f"Filepath '{self.filepath}' does not exist",
                code=WorkflowErrorCode.INVALID_INPUTS,
            )

        return code, os.path.join(os.path.dirname(root), self.filepath)



from typing import Any, Literal, Union

from vellum.client.core.pydantic_utilities import UniversalBaseModel


class BaseStateDelta(UniversalBaseModel):
    name: str


class SetStateDelta(BaseStateDelta):
    delta: Any
    delta_type: Literal["SET_STATE"] = "SET_STATE"


class AppendStateDelta(BaseStateDelta):
    delta: Any
    delta_type: Literal["APPEND_STATE"] = "APPEND_STATE"


StateDelta = Union[SetStateDelta, AppendStateDelta]



"""File upload utilities for uploading files to Vellum."""

import base64
from io import BytesIO
import logging
import re
from typing import TYPE_CHECKING, Optional

import requests

from vellum.client.core.api_error import ApiError
from vellum.client.core.file import File
from vellum.utils.files.constants import BASE64_DATA_URL_PATTERN, URL_PATTERN, VELLUM_FILE_SRC_PATTERN
from vellum.utils.files.exceptions import FileNotFoundError, FileRetrievalError, InvalidFileSourceError
from vellum.utils.files.extensions import ensure_filename_with_extension
from vellum.utils.files.types import VellumFileTypes

if TYPE_CHECKING:
    from vellum.client import Vellum as VellumClient

logger = logging.getLogger(__name__)


def upload_vellum_file(
    vellum_file: VellumFileTypes,
    *,
    filename: Optional[str] = None,
    vellum_client: Optional["VellumClient"] = None,
) -> VellumFileTypes:
    """
    Upload a file to Vellum and return a new VellumFile with the uploaded source.

    This function takes any VellumFile object (with a src that could be a base64 data URL,
    HTTP/HTTPS URL, or existing vellum:uploaded-file: identifier), downloads its content,
    and uploads it to Vellum's storage.

    Args:
        vellum_file: A VellumDocument, VellumImage, VellumAudio, or VellumVideo instance
        filename: Optional filename to use when uploading. If not provided, defaults to "file"
                 with an appropriate extension inferred from the MIME type. If provided without
                 an extension, the extension will be automatically added based on the MIME type.
        vellum_client: An optional Vellum client instance. If not provided, a default client will be created.

    Returns:
        VellumFileTypes: A new VellumFile of the same type with the vellum:uploaded-file:{id} source

    Raises:
        InvalidFileSourceError: If the file source format is invalid
        FileRetrievalError: If the file cannot be retrieved from its source
        FileNotFoundError: If the file is not found at its source

    Example:
        ```python
        from vellum import VellumDocument, VellumImage
        from vellum.utils.files import upload_vellum_file

        # Upload a document from a URL
        doc = VellumDocument(src="https://example.com/doc.pdf")
        uploaded_doc = upload_vellum_file(doc, filename="my_document.pdf")

        # Upload a base64-encoded image (filename automatically inferred as "file.png")
        image = VellumImage(src="data:image/png;base64,iVBORw0KGgo...")
        uploaded_image = upload_vellum_file(image)

        # Upload with filename without extension (extension automatically added)
        doc = VellumDocument(src="data:application/pdf;base64,...")
        uploaded_doc = upload_vellum_file(doc, filename="report")  # becomes "report.pdf"
        ```
    """
    src = vellum_file.src

    # Case 1: Vellum Uploaded File (already uploaded, just return as-is)
    match = re.match(VELLUM_FILE_SRC_PATTERN, src, re.IGNORECASE)
    if match:
        vellum_uploaded_file_id = match.group(1)
        logger.info(
            "File already uploaded to Vellum, returning unchanged",
            extra={"vellum_uploaded_file_id": vellum_uploaded_file_id},
        )
        return vellum_file

    if vellum_client is None:
        from vellum.utils.vellum_client import create_vellum_client

        vellum_client = create_vellum_client()

    # Case 2: Base64 Data URL
    data_url_match = re.match(BASE64_DATA_URL_PATTERN, src)
    if data_url_match:
        mime_type = data_url_match.group(1) or "application/octet-stream"
        base64_content = data_url_match.group(3)
        decoded = base64.b64decode(base64_content)

        # Ensure filename has appropriate extension
        resolved_filename = ensure_filename_with_extension(filename, mime_type)
        file_content: File = (resolved_filename, BytesIO(decoded), mime_type)

        try:
            uploaded_file = vellum_client.uploaded_files.create(file=file_content)
        except ApiError as e:
            raise FileRetrievalError(f"Failed to upload file to Vellum: {str(e)}") from e

        new_src = f"vellum:uploaded-file:{uploaded_file.id}"
        return _copy_with_new_src(vellum_file, new_src)

    # Case 3: Direct URL
    if not re.match(URL_PATTERN, src):
        raise InvalidFileSourceError(
            f"Invalid file source: {src}. "
            "Expected a base64 data URL (data:...;base64,...), "
            "a Vellum uploaded file ID (vellum:uploaded-file:...), "
            "or an HTTP/HTTPS URL."
        )

    file_url = src

    # Download from URL
    try:
        response = requests.get(file_url, stream=True)
        response.raise_for_status()
    except requests.HTTPError as e:
        if e.response.status_code == 404:
            raise FileNotFoundError(f"File not found at URL: {file_url}") from e
        raise FileRetrievalError(f"Failed to retrieve file from URL: {file_url}") from e
    except requests.RequestException as e:
        raise FileRetrievalError(f"Network error while retrieving file: {file_url}") from e

    # Read content and get content type
    content = response.content
    content_type = response.headers.get("content-type", "application/octet-stream")

    # Ensure filename has appropriate extension
    resolved_filename = ensure_filename_with_extension(filename, content_type)
    file_content = (resolved_filename, BytesIO(content), content_type)

    try:
        uploaded_file = vellum_client.uploaded_files.create(file=file_content)
    except ApiError as e:
        raise FileRetrievalError("Failed to upload file to Vellum") from e

    new_src = f"vellum:uploaded-file:{uploaded_file.id}"
    return _copy_with_new_src(vellum_file, new_src)


def _copy_with_new_src(vellum_file: VellumFileTypes, new_src: str) -> VellumFileTypes:
    """Helper to create a copy of a VellumFile with an updated src."""
    # Use model_copy for Pydantic v2, copy for v1
    if hasattr(vellum_file, "model_copy"):
        return vellum_file.model_copy(update={"src": new_src})
    else:
        return vellum_file.copy(update={"src": new_src})



import pytest
from uuid import UUID

from vellum.workflows.constants import AuthorizationType, VellumIntegrationProviderType
from vellum.workflows.references.environment_variable import EnvironmentVariableReference
from vellum.workflows.types.definition import (
    ComposioToolDefinition,
    DeploymentDefinition,
    MCPServer,
    MCPToolDefinition,
    VellumIntegrationToolDefinition,
)


@pytest.mark.parametrize(
    "deployment_value, expected_deployment_id, expected_deployment_name",
    [
        # Valid UUID string
        (
            "57f09beb-b463-40e0-bf9e-c972e664352f",
            UUID("57f09beb-b463-40e0-bf9e-c972e664352f"),
            None,
        ),
        # Name string
        (
            "tool-calling-subworkflow",
            None,
            "tool-calling-subworkflow",
        ),
    ],
    ids=[
        "valid_uuid",
        "valid_name",
    ],
)
def test_deployment_definition(deployment_value, expected_deployment_id, expected_deployment_name):
    """Test that DeploymentDefinition properties correctly identify and extract UUID vs name."""
    deployment = DeploymentDefinition(deployment=deployment_value)

    assert deployment.deployment_id == expected_deployment_id
    assert deployment.deployment_name == expected_deployment_name


def test_composio_tool_definition_creation():
    """Test that ComposioToolDefinition can be created with required fields."""
    composio_tool = ComposioToolDefinition(
        toolkit="GITHUB",
        action="GITHUB_CREATE_AN_ISSUE",
        description="Create a new issue in a GitHub repository",
        user_id=None,
    )

    assert composio_tool.toolkit == "GITHUB"
    assert composio_tool.action == "GITHUB_CREATE_AN_ISSUE"
    assert composio_tool.description == "Create a new issue in a GitHub repository"
    assert composio_tool.user_id is None
    assert composio_tool.name == "github_create_an_issue"


def test_mcp_tool_definition_creation_bearer_token():
    """Test that MCPToolDefinition can be created with required fields."""
    mcp_tool = MCPToolDefinition(
        name="create_repository",
        server=MCPServer(
            name="github",
            url="https://api.githubcopilot.com/mcp/",
            authorization_type=AuthorizationType.BEARER_TOKEN,
            bearer_token_value=EnvironmentVariableReference(name="GITHUB_PERSONAL_ACCESS_TOKEN"),
        ),
        parameters={
            "type": "object",
            "properties": {
                "repository_name": {"type": "string", "description": "Repository name"},
                "description": {"type": "string", "description": "Repository description"},
            },
            "required": ["repository_name"],
        },
    )

    assert mcp_tool.name == "create_repository"
    assert mcp_tool.server.name == "github"
    assert mcp_tool.server.url == "https://api.githubcopilot.com/mcp/"
    assert mcp_tool.server.authorization_type == AuthorizationType.BEARER_TOKEN
    assert mcp_tool.parameters == {
        "type": "object",
        "properties": {
            "repository_name": {"type": "string", "description": "Repository name"},
            "description": {"type": "string", "description": "Repository description"},
        },
        "required": ["repository_name"],
    }

    assert isinstance(mcp_tool.server.bearer_token_value, EnvironmentVariableReference)
    assert mcp_tool.server.bearer_token_value.name == "GITHUB_PERSONAL_ACCESS_TOKEN"


def test_mcp_tool_definition_creation_api_key():
    """Test that MCPToolDefinition can be created with required fields."""
    mcp_tool = MCPToolDefinition(
        name="create_repository",
        server=MCPServer(
            name="github",
            url="https://api.githubcopilot.com/mcp/",
            authorization_type=AuthorizationType.API_KEY,
            api_key_header_key="Authorization",
            api_key_header_value=EnvironmentVariableReference(name="GITHUB_PERSONAL_ACCESS_TOKEN"),
        ),
        parameters={
            "type": "object",
            "properties": {
                "repository_name": {"type": "string", "description": "Repository name"},
                "description": {"type": "string", "description": "Repository description"},
            },
            "required": ["repository_name"],
        },
    )

    assert mcp_tool.name == "create_repository"
    assert mcp_tool.server.name == "github"
    assert mcp_tool.server.url == "https://api.githubcopilot.com/mcp/"
    assert mcp_tool.server.authorization_type == AuthorizationType.API_KEY
    assert mcp_tool.server.api_key_header_key == "Authorization"
    assert isinstance(mcp_tool.server.api_key_header_value, EnvironmentVariableReference)
    assert mcp_tool.server.api_key_header_value.name == "GITHUB_PERSONAL_ACCESS_TOKEN"
    assert mcp_tool.parameters == {
        "type": "object",
        "properties": {
            "repository_name": {"type": "string", "description": "Repository name"},
            "description": {"type": "string", "description": "Repository description"},
        },
        "required": ["repository_name"],
    }


def test_mcp_tool_definition_creation_no_authorization():
    """Test that MCPToolDefinition can be created with no authorization."""
    mcp_tool = MCPToolDefinition(
        name="create_repository",
        server=MCPServer(
            name="github",
            url="https://api.githubcopilot.com/mcp/",
        ),
    )

    assert mcp_tool.name == "create_repository"
    assert mcp_tool.server.name == "github"
    assert mcp_tool.server.url == "https://api.githubcopilot.com/mcp/"
    assert mcp_tool.server.authorization_type is None
    assert mcp_tool.server.bearer_token_value is None
    assert mcp_tool.server.api_key_header_key is None
    assert mcp_tool.server.api_key_header_value is None


def test_vellum_integration_tool_definition_creation():
    """Test that VellumIntegrationToolDefinition can be created with required fields."""
    vellum_tool = VellumIntegrationToolDefinition(
        provider=VellumIntegrationProviderType.COMPOSIO,
        integration_name="GITHUB",
        name="create_issue",
        description="Create a new issue in a GitHub repository",
    )

    assert vellum_tool.type == "VELLUM_INTEGRATION"
    assert vellum_tool.provider == VellumIntegrationProviderType.COMPOSIO
    assert vellum_tool.integration_name == "GITHUB"
    assert vellum_tool.name == "create_issue"
    assert vellum_tool.description == "Create a new issue in a GitHub repository"


def test_vellum_integration_tool_definition_with_different_provider():
    """Test VellumIntegrationToolDefinition with a different provider."""
    vellum_tool = VellumIntegrationToolDefinition(
        provider=VellumIntegrationProviderType.COMPOSIO,
        integration_name="SLACK",
        name="send_message",
        description="Send a message to a Slack channel",
    )

    assert vellum_tool.type == "VELLUM_INTEGRATION"
    assert vellum_tool.provider == VellumIntegrationProviderType.COMPOSIO
    assert vellum_tool.integration_name == "SLACK"
    assert vellum_tool.name == "send_message"
    assert vellum_tool.description == "Send a message to a Slack channel"



"""Streaming utilities for reading Vellum files."""

import base64
from contextlib import contextmanager
from io import BytesIO
import re
from typing import TYPE_CHECKING, Generator, Iterator, Optional

import requests

from vellum.client.core.api_error import ApiError
from vellum.utils.files.constants import BASE64_DATA_URL_PATTERN, URL_PATTERN, VELLUM_FILE_SRC_PATTERN
from vellum.utils.files.exceptions import FileNotFoundError, FileRetrievalError, InvalidFileSourceError
from vellum.utils.files.types import VellumFileTypes

if TYPE_CHECKING:
    from vellum.client import Vellum as VellumClient


@contextmanager
def stream_vellum_file(
    vellum_file: VellumFileTypes,
    *,
    chunk_size: int = 8192,
    vellum_client: Optional["VellumClient"] = None,
) -> Generator[Iterator[bytes], None, None]:
    """
    Stream the file content in chunks using a context manager.

    This is the recommended way to read large files, as it avoids loading
    the entire file into memory at once.

    Args:
        vellum_file: A VellumDocument, VellumImage, VellumAudio, or VellumVideo instance
        chunk_size: Size of chunks to yield in bytes (default 8KB)
        vellum_client: An optional Vellum client instance. If not provided, a default client will be created.

    Yields:
        Iterator[bytes]: Chunks of file content

    Example:
        ```python
        from vellum import VellumVideo
        from vellum.utils.files import stream_vellum_file

        # Stream a large video file
        video = VellumVideo(src="https://example.com/video.mp4")
        with stream_vellum_file(video, chunk_size=10*1024*1024) as chunks:  # 10MB chunks
            with open('output.mp4', 'wb') as f:
                for chunk in chunks:
                    f.write(chunk)

        # Calculate hash without loading entire file
        import hashlib
        hasher = hashlib.sha256()
        with stream_vellum_file(vellum_file) as chunks:
            for chunk in chunks:
                hasher.update(chunk)
        hash_value = hasher.hexdigest()
        ```
    """
    src = vellum_file.src

    # Case 1: Base64 Data URL
    # Note: Base64 content is already in memory, but we still provide
    # a streaming interface for consistency
    data_url_match = re.match(BASE64_DATA_URL_PATTERN, src)
    if data_url_match:
        base64_content = data_url_match.group(3)
        decoded = base64.b64decode(base64_content)
        stream = BytesIO(decoded)
        try:
            yield _chunk_iterator(stream, chunk_size)
        finally:
            stream.close()
        return

    file_url: str

    # Case 2: Vellum Uploaded File
    match = re.match(VELLUM_FILE_SRC_PATTERN, src, re.IGNORECASE)
    if match:
        vellum_uploaded_file_id = match.group(1)
        if vellum_client is None:
            from vellum.utils.vellum_client import create_vellum_client

            vellum_client = create_vellum_client()

        try:
            uploaded_file = vellum_client.uploaded_files.retrieve(vellum_uploaded_file_id)
        except ApiError as e:
            if e.status_code == 404:
                raise FileNotFoundError(f"Uploaded file not found: {vellum_uploaded_file_id}") from e
            raise FileRetrievalError(f"Failed to retrieve uploaded file: {vellum_uploaded_file_id}") from e

        if not uploaded_file.file_url:
            raise FileRetrievalError(f"Uploaded file has no accessible URL: {vellum_uploaded_file_id}")

        file_url = uploaded_file.file_url

    # Case 3: Direct URL
    elif re.match(URL_PATTERN, src):
        file_url = src
    else:
        raise InvalidFileSourceError(
            f"Invalid file source: {src}. "
            "Expected a base64 data URL (data:...;base64,...), "
            "a Vellum uploaded file ID (vellum:uploaded-file:...), "
            "or an HTTP/HTTPS URL."
        )

    # Stream from URL
    try:
        response = requests.get(file_url, stream=True)
        response.raise_for_status()
    except requests.HTTPError as e:
        if e.response.status_code == 404:
            raise FileNotFoundError(f"File not found at URL: {file_url}") from e
        raise FileRetrievalError(f"Failed to retrieve file from URL: {file_url}") from e
    except requests.RequestException as e:
        raise FileRetrievalError(f"Network error while retrieving file: {file_url}") from e

    try:
        yield response.iter_content(chunk_size=chunk_size)
    finally:
        response.close()


def _chunk_iterator(stream: BytesIO, chunk_size: int) -> Iterator[bytes]:
    """Helper to read a BytesIO stream in chunks."""
    while True:
        chunk = stream.read(chunk_size)
        if not chunk:
            break
        yield chunk



# This file was auto-generated by Fern from our API Definition.

import typing

from .plain_text_prompt_block import PlainTextPromptBlock
from .variable_prompt_block import VariablePromptBlock

RichTextChildBlock = typing.Union[VariablePromptBlock, PlainTextPromptBlock]



from abc import ABC, abstractmethod
from uuid import UUID
from typing import TYPE_CHECKING, Iterator, Optional, Type, Union

from vellum.workflows.events.workflow import WorkflowEvent
from vellum.workflows.resolvers.types import LoadStateResult
from vellum.workflows.state.base import BaseState

if TYPE_CHECKING:
    from vellum.workflows.state.context import WorkflowContext
    from vellum.workflows.workflows.base import BaseWorkflow


class BaseWorkflowResolver(ABC):
    def __init__(self):
        self._context: Optional["WorkflowContext"] = None
        self._workflow_class: Optional[Type["BaseWorkflow"]] = None

    def register_workflow_instance(self, workflow_instance: "BaseWorkflow") -> None:
        self._workflow_class = type(workflow_instance)
        self._context = workflow_instance.context

    @abstractmethod
    def get_latest_execution_events(self) -> Iterator[WorkflowEvent]:
        pass

    @abstractmethod
    def get_state_snapshot_history(self) -> Iterator[BaseState]:
        pass

    @abstractmethod
    def load_state(self, previous_execution_id: Optional[Union[UUID, str]] = None) -> Optional[LoadStateResult]:
        pass



from deepdiff import DeepDiff

from vellum_ee.workflows.display.workflows.get_vellum_workflow_display_class import get_workflow_display

from tests.workflows.basic_tool_calling_node_parent_input.workflow import BasicToolCallingNodeParentInputWorkflow


def test_serialize_workflow():
    # GIVEN a Workflow that uses a generic node
    # WHEN we serialize it
    workflow_display = get_workflow_display(workflow_class=BasicToolCallingNodeParentInputWorkflow)

    serialized_workflow: dict = workflow_display.serialize()
    # THEN we should get a serialized representation of the Workflow
    assert serialized_workflow.keys() == {
        "workflow_raw_data",
        "input_variables",
        "state_variables",
        "output_variables",
    }

    # AND its input variables should be what we expect
    input_variables = serialized_workflow["input_variables"]
    assert len(input_variables) == 1

    # AND its output variables should be what we expect
    output_variables = serialized_workflow["output_variables"]
    assert len(output_variables) == 2
    assert not DeepDiff(
        [
            {"id": "e2e36cfc-cf24-42fd-ba8f-cce39c53d47b", "key": "text", "type": "STRING"},
            {"id": "08ca9519-e421-47e8-a42d-44f49f6aab16", "key": "chat_history", "type": "CHAT_HISTORY"},
        ],
        output_variables,
        ignore_order=True,
    )

    # AND its raw data should be what we expect
    workflow_raw_data = serialized_workflow["workflow_raw_data"]
    tool_calling_node = workflow_raw_data["nodes"][2]

    attributes = tool_calling_node["attributes"]
    function_attributes = next(attribute for attribute in attributes if attribute["name"] == "functions")
    assert function_attributes == {
        "id": "e2695720-919b-497e-8d01-ae6aa705f7bb",
        "name": "functions",
        "value": {
            "type": "CONSTANT_VALUE",
            "value": {
                "type": "JSON",
                "value": [
                    {
                        "type": "CODE_EXECUTION",
                        "name": "get_string",
                        "description": "\n    Get a string with the parent input, dummy input, and the populated input.\n    ",  # noqa: E501
                        "definition": {
                            "state": None,
                            "cache_config": None,
                            "name": "get_string",
                            "description": "\n    Get a string with the parent input, dummy input, and the populated input.\n    ",  # noqa: E501
                            "parameters": {
                                "type": "object",
                                "properties": {"populated_input": {"type": "string"}},
                                "required": ["populated_input"],
                            },
                            "inputs": {
                                "parent_input": {
                                    "type": "WORKFLOW_INPUT",
                                    "input_variable_id": "4bf1f0e7-76c6-4204-9f8c-bd9c3b73a8db",
                                },
                                "dummy_input": {
                                    "type": "NODE_OUTPUT",
                                    "node_id": "72f78142-e0a2-40a9-ae70-0230ccf3b503",
                                    "node_output_id": "6e639661-d0dc-4586-9393-e64e95e0d3ef",
                                },
                                "constant_input": {
                                    "type": "CONSTANT_VALUE",
                                    "value": {"type": "STRING", "value": "constant_input"},
                                },
                            },
                            "forced": None,
                            "strict": None,
                        },
                        "src": 'from vellum.workflows.utils.functions import use_tool_inputs\n\nfrom .inputs import ParentInputs\nfrom .nodes.dummy_node import DummyNode\n\n\n@use_tool_inputs(\n    parent_input=ParentInputs.parent_input,\n    dummy_input=DummyNode.Outputs.text,\n    constant_input="constant_input",\n)\ndef get_string(parent_input: str, dummy_input: str, constant_input: str, populated_input: str) -> str:\n    """\n    Get a string with the parent input, dummy input, and the populated input.\n    """\n    return f"parent input: {parent_input}, dummy input: {dummy_input}, constant input: {constant_input}, populated input: {populated_input}"  # noqa: E501\n',  # noqa: E501
                    }
                ],
            },
        },
    }



from datetime import datetime
from uuid import uuid4

from deepdiff import DeepDiff

from vellum import WorkflowDeploymentRead
from vellum_ee.workflows.display.workflows.get_vellum_workflow_display_class import get_workflow_display

from tests.workflows.basic_subworkflow_deployment.workflow import BasicSubworkflowDeploymentWorkflow


def test_serialize_workflow(vellum_client):
    # GIVEN a Workflow with stubbed out API calls
    deployment = WorkflowDeploymentRead(
        id=str(uuid4()),
        created=datetime.now(),
        label="Example Subworkflow Deployment",
        name="example_subworkflow_deployment",
        input_variables=[],
        output_variables=[],
        last_deployed_on=datetime.now(),
        last_deployed_history_item_id=str(uuid4()),
    )
    vellum_client.workflow_deployments.retrieve.return_value = deployment

    # WHEN we serialize it
    workflow_display = get_workflow_display(workflow_class=BasicSubworkflowDeploymentWorkflow)
    serialized_workflow: dict = workflow_display.serialize()

    # THEN we should get a serialized representation of the Workflow
    assert serialized_workflow.keys() == {
        "workflow_raw_data",
        "input_variables",
        "state_variables",
        "output_variables",
    }

    # AND its input variables should be what we expect
    input_variables = serialized_workflow["input_variables"]
    assert len(input_variables) == 2
    assert not DeepDiff(
        [
            {
                "id": "693cc9a5-8d74-4a58-bdcf-2b4989cdf250",
                "key": "city",
                "type": "STRING",
                "required": True,
                "default": None,
                "extensions": {"color": None},
            },
            {
                "id": "19a78824-9a98-4ae8-a1fc-61f81a422a17",
                "key": "date",
                "type": "STRING",
                "required": True,
                "default": None,
                "extensions": {"color": None},
            },
        ],
        input_variables,
        ignore_order=True,
    )

    # AND its output variables should be what we expect
    output_variables = serialized_workflow["output_variables"]
    assert len(output_variables) == 2
    assert not DeepDiff(
        [
            {
                "id": "3f487916-126f-4d6c-95b4-fa72d875b793",
                "key": "temperature",
                "type": "NUMBER",
            },
            {
                "id": "45d53a1e-26e8-4c43-a010-80d141acc249",
                "key": "reasoning",
                "type": "STRING",
            },
        ],
        output_variables,
        ignore_order=True,
    )

    # AND its raw data should be what we expect
    workflow_raw_data = serialized_workflow["workflow_raw_data"]
    assert len(workflow_raw_data["edges"]) == 3
    assert len(workflow_raw_data["nodes"]) == 4

    # AND each node should be serialized correctly
    entrypoint_node = workflow_raw_data["nodes"][0]
    assert entrypoint_node == {
        "id": "f0eea82b-39cc-44e3-9c0d-12205ed5652c",
        "type": "ENTRYPOINT",
        "base": None,
        "definition": None,
        "inputs": [],
        "data": {
            "label": "Entrypoint Node",
            "source_handle_id": "13d9eb34-aecb-496d-9e57-d5e786b0bc7c",
        },
        "display_data": {
            "position": {"x": 0.0, "y": -50.0},
        },
    }

    subworkflow_node = workflow_raw_data["nodes"][1]
    assert subworkflow_node == {
        "id": "bb98a2c4-c9a7-4c39-8f31-dc7961dc9996",
        "type": "SUBWORKFLOW",
        "inputs": [
            {
                "id": "8107cec2-8215-4730-b52c-859e87a1c116",
                "key": "city",
                "value": {
                    "rules": [
                        {
                            "type": "INPUT_VARIABLE",
                            "data": {"input_variable_id": "693cc9a5-8d74-4a58-bdcf-2b4989cdf250"},
                        }
                    ],
                    "combinator": "OR",
                },
            },
            {
                "id": "3487e51a-e7fe-4b2c-a1f9-f72c83a329db",
                "key": "date",
                "value": {
                    "rules": [
                        {
                            "type": "INPUT_VARIABLE",
                            "data": {"input_variable_id": "19a78824-9a98-4ae8-a1fc-61f81a422a17"},
                        }
                    ],
                    "combinator": "OR",
                },
            },
        ],
        "data": {
            "label": "Example Subworkflow Deployment Node",
            "error_output_id": None,
            "source_handle_id": "ff99bf0c-c239-4b8b-8ac1-483b134f94f4",
            "target_handle_id": "d6194ccf-d31b-4846-8e24-3e189d84351a",
            "variant": "DEPLOYMENT",
            "workflow_deployment_id": deployment.id,
            "release_tag": "LATEST",
        },
        "display_data": {"position": {"x": 200.0, "y": -50.0}},
        "base": {
            "module": ["vellum", "workflows", "nodes", "displayable", "subworkflow_deployment_node", "node"],
            "name": "SubworkflowDeploymentNode",
        },
        "definition": {
            "module": ["tests", "workflows", "basic_subworkflow_deployment", "workflow"],
            "name": "ExampleSubworkflowDeploymentNode",
        },
        "trigger": {
            "id": "d6194ccf-d31b-4846-8e24-3e189d84351a",
            "merge_behavior": "AWAIT_ANY",
        },
        "ports": [{"id": "ff99bf0c-c239-4b8b-8ac1-483b134f94f4", "name": "default", "type": "DEFAULT"}],
    }

    assert not DeepDiff(
        [
            {
                "id": "18170041-1a70-4836-9fa0-adceba2a1f4f",
                "type": "TERMINAL",
                "base": {
                    "module": [
                        "vellum",
                        "workflows",
                        "nodes",
                        "displayable",
                        "final_output_node",
                        "node",
                    ],
                    "name": "FinalOutputNode",
                },
                "definition": None,
                "data": {
                    "label": "Final Output",
                    "name": "temperature",
                    "target_handle_id": "23117248-df28-4519-bebc-abcb24f966b3",
                    "output_id": "3f487916-126f-4d6c-95b4-fa72d875b793",
                    "output_type": "NUMBER",
                    "node_input_id": "6f4955d8-8a3f-4db7-8293-7affc5877dcd",
                },
                "inputs": [
                    {
                        "id": "6f4955d8-8a3f-4db7-8293-7affc5877dcd",
                        "key": "node_input",
                        "value": {
                            "rules": [
                                {
                                    "type": "NODE_OUTPUT",
                                    "data": {
                                        "node_id": "bb98a2c4-c9a7-4c39-8f31-dc7961dc9996",
                                        "output_id": "d901cbed-9905-488c-be62-e2668f85438f",
                                    },
                                }
                            ],
                            "combinator": "OR",
                        },
                    }
                ],
                "display_data": {"position": {"x": 400.0, "y": -175.0}},
            },
            {
                "id": "94afd0ac-1ec4-486b-a6fb-fa1ec7029d19",
                "type": "TERMINAL",
                "base": {
                    "module": [
                        "vellum",
                        "workflows",
                        "nodes",
                        "displayable",
                        "final_output_node",
                        "node",
                    ],
                    "name": "FinalOutputNode",
                },
                "definition": None,
                "data": {
                    "label": "Final Output",
                    "name": "reasoning",
                    "target_handle_id": "c3aeba92-4faf-4814-9842-eec7436ee555",
                    "output_id": "45d53a1e-26e8-4c43-a010-80d141acc249",
                    "output_type": "STRING",
                    "node_input_id": "74337307-3fcb-42c5-9aed-98bd4a79caef",
                },
                "inputs": [
                    {
                        "id": "74337307-3fcb-42c5-9aed-98bd4a79caef",
                        "key": "node_input",
                        "value": {
                            "rules": [
                                {
                                    "type": "NODE_OUTPUT",
                                    "data": {
                                        "node_id": "bb98a2c4-c9a7-4c39-8f31-dc7961dc9996",
                                        "output_id": "68de689c-fe8a-4189-b7d0-82c620ac30f9",
                                    },
                                }
                            ],
                            "combinator": "OR",
                        },
                    }
                ],
                "display_data": {"position": {"x": 400.0, "y": 75.0}},
            },
        ],
        workflow_raw_data["nodes"][2:],
        ignore_order=True,
    )

    # AND each edge should be serialized correctly
    serialized_edges = workflow_raw_data["edges"]
    assert not DeepDiff(
        [
            {
                "id": "4a08ef93-c0f4-4a3a-9a24-5d79ee4d85fc",
                "source_handle_id": "13d9eb34-aecb-496d-9e57-d5e786b0bc7c",
                "source_node_id": "f0eea82b-39cc-44e3-9c0d-12205ed5652c",
                "target_handle_id": "d6194ccf-d31b-4846-8e24-3e189d84351a",
                "target_node_id": "bb98a2c4-c9a7-4c39-8f31-dc7961dc9996",
                "type": "DEFAULT",
            },
            {
                "id": "69933897-e91e-4c6c-9ba3-ed3e3c265c73",
                "source_handle_id": "ff99bf0c-c239-4b8b-8ac1-483b134f94f4",
                "source_node_id": "bb98a2c4-c9a7-4c39-8f31-dc7961dc9996",
                "target_handle_id": "c3aeba92-4faf-4814-9842-eec7436ee555",
                "target_node_id": "94afd0ac-1ec4-486b-a6fb-fa1ec7029d19",
                "type": "DEFAULT",
            },
            {
                "id": "86a9af31-f78e-45ac-b170-f66bbba98f9d",
                "source_handle_id": "ff99bf0c-c239-4b8b-8ac1-483b134f94f4",
                "source_node_id": "bb98a2c4-c9a7-4c39-8f31-dc7961dc9996",
                "target_handle_id": "23117248-df28-4519-bebc-abcb24f966b3",
                "target_node_id": "18170041-1a70-4836-9fa0-adceba2a1f4f",
                "type": "DEFAULT",
            },
        ],
        serialized_edges,
        ignore_order=True,
    )

    # AND the display data should be what we expect
    display_data = workflow_raw_data["display_data"]
    assert display_data == {
        "viewport": {
            "x": 0.0,
            "y": 0.0,
            "zoom": 1.0,
        }
    }

    # AND the definition should be what we expect
    definition = workflow_raw_data["definition"]
    assert definition == {
        "name": "BasicSubworkflowDeploymentWorkflow",
        "module": [
            "tests",
            "workflows",
            "basic_subworkflow_deployment",
            "workflow",
        ],
    }



import pytest
from unittest import mock
from uuid import uuid4

from vellum import Integration
from vellum.client.types.components_schemas_composio_tool_definition import ComponentsSchemasComposioToolDefinition
from vellum.workflows.constants import VellumIntegrationProviderType
from vellum.workflows.exceptions import NodeException
from vellum.workflows.integrations.vellum_integration_service import VellumIntegrationService
from vellum.workflows.types.definition import VellumIntegrationToolDetails


def test_vellum_integration_service_get_tool_definition_success(vellum_client):
    """Test that tool definitions are successfully retrieved from Vellum API"""
    # GIVEN a mock client configured to return a tool definition
    mock_client = vellum_client
    tool_definition_response = ComponentsSchemasComposioToolDefinition(
        integration=Integration(
            id=str(uuid4()),
            provider="COMPOSIO",
            name="GITHUB",
        ),
        label="GITHUB_CREATE_AN_ISSUE",
        name="GITHUB_CREATE_AN_ISSUE",
        description="Create a new issue in a GitHub repository",
        input_parameters={
            "type": "object",
            "properties": {
                "repo": {"type": "string", "description": "Repository name"},
                "title": {"type": "string", "description": "Issue title"},
                "body": {"type": "string", "description": "Issue body"},
            },
            "required": ["repo", "title"],
        },
        output_parameters={},
    )
    mock_client.integrations.retrieve_integration_tool_definition.return_value = tool_definition_response

    # WHEN we request a tool definition
    service = VellumIntegrationService(client=mock_client)
    result = service.get_tool_definition(
        integration="GITHUB",
        provider="COMPOSIO",
        tool_name="GITHUB_CREATE_AN_ISSUE",
    )

    # THEN the tool definition should be returned with all expected fields
    assert isinstance(result, VellumIntegrationToolDetails)
    assert result.name == "GITHUB_CREATE_AN_ISSUE"
    assert result.description == "Create a new issue in a GitHub repository"
    assert result.provider == VellumIntegrationProviderType.COMPOSIO
    # Parameters should now be included in the tool details
    assert result.parameters is not None
    assert result.parameters["type"] == "object"
    assert "properties" in result.parameters
    assert "repo" in result.parameters["properties"]
    assert "title" in result.parameters["properties"]

    # AND the API should have been called with the correct parameters
    mock_client.integrations.retrieve_integration_tool_definition.assert_called_once_with(
        integration_name="GITHUB",
        integration_provider="COMPOSIO",
        tool_name="GITHUB_CREATE_AN_ISSUE",
    )


def test_vellum_integration_service_get_tool_definition_api_error(vellum_client):
    """Test that API errors are properly handled when retrieving tool definitions"""
    # GIVEN a mock client configured to raise an exception
    mock_client = vellum_client
    mock_client.integrations = mock.MagicMock()
    mock_client.integrations.retrieve_integration_tool_definition.side_effect = Exception("Tool not found")

    # WHEN we attempt to get a tool definition for an invalid tool
    service = VellumIntegrationService(client=mock_client)

    # THEN it should raise a NodeException with appropriate error message
    with pytest.raises(NodeException) as exc_info:
        service.get_tool_definition(
            integration="GITHUB",
            provider="COMPOSIO",
            tool_name="INVALID_TOOL",
        )

    assert "Failed to retrieve tool definition" in str(exc_info.value)
    assert "Tool not found" in str(exc_info.value)


def test_vellum_integration_service_execute_tool_success(vellum_client):
    """Test that tools are successfully executed via Vellum API"""
    # GIVEN a mock client configured to return a successful response
    mock_client = vellum_client
    mock_client.integrations = mock.MagicMock()

    mock_response = mock.MagicMock()
    mock_response.data = {
        "success": True,
        "issue_id": 123,
        "issue_url": "https://github.com/user/repo/issues/123",
    }
    mock_client.integrations.execute_integration_tool.return_value = mock_response

    # WHEN we execute a tool with valid arguments
    service = VellumIntegrationService(client=mock_client)
    result = service.execute_tool(
        integration="GITHUB",
        provider="COMPOSIO",
        tool_name="GITHUB_CREATE_AN_ISSUE",
        arguments={
            "repo": "user/repo",
            "title": "Test Issue",
            "body": "Test body",
        },
    )

    # THEN the execution result should contain expected data
    assert result["success"] is True
    assert result["issue_id"] == 123
    assert result["issue_url"] == "https://github.com/user/repo/issues/123"

    # AND the API should have been called with correct parameters
    mock_client.integrations.execute_integration_tool.assert_called_once_with(
        integration_name="GITHUB",
        integration_provider="COMPOSIO",
        tool_name="GITHUB_CREATE_AN_ISSUE",
        arguments={
            "repo": "user/repo",
            "title": "Test Issue",
            "body": "Test body",
        },
    )


def test_vellum_integration_service_execute_tool_api_error(vellum_client):
    """Test that execution errors are properly handled"""
    # GIVEN a mock client configured to raise an exception
    mock_client = vellum_client
    mock_client.integrations = mock.MagicMock()
    mock_client.integrations.execute_integration_tool.side_effect = Exception("Authentication failed")

    # WHEN we attempt to execute a tool that encounters an error
    service = VellumIntegrationService(client=mock_client)

    # THEN it should raise a NodeException with appropriate error message
    with pytest.raises(NodeException) as exc_info:
        service.execute_tool(
            integration="GITHUB",
            provider="COMPOSIO",
            tool_name="GITHUB_CREATE_AN_ISSUE",
            arguments={"repo": "user/repo"},
        )

    assert "Failed to execute tool" in str(exc_info.value)
    assert "Authentication failed" in str(exc_info.value)


def test_vellum_integration_service_execute_tool_empty_response(vellum_client):
    """Test that empty response data is handled gracefully"""
    # GIVEN a mock client configured to return an empty response
    mock_client = vellum_client
    mock_client.integrations = mock.MagicMock()

    mock_response = mock.MagicMock()
    mock_response.data = {}
    mock_client.integrations.execute_integration_tool.return_value = mock_response

    # WHEN we execute a tool that returns empty data
    service = VellumIntegrationService(client=mock_client)
    result = service.execute_tool(
        integration="SLACK",
        provider="COMPOSIO",
        tool_name="SLACK_SEND_MESSAGE",
        arguments={
            "channel": "#general",
            "message": "Hello, world!",
        },
    )

    # THEN an empty dictionary should be returned without errors
    assert result == {}


def test_vellum_integration_service_multiple_tool_executions(vellum_client):
    """Test that the service handles multiple sequential tool executions"""
    # GIVEN a mock client configured to return different responses for multiple calls
    mock_client = vellum_client
    mock_client.integrations = mock.MagicMock()

    responses = [
        mock.MagicMock(data={"result": "first"}),
        mock.MagicMock(data={"result": "second"}),
    ]
    mock_client.integrations.execute_integration_tool.side_effect = responses

    # WHEN we execute multiple tools in sequence
    service = VellumIntegrationService(client=mock_client)

    result1 = service.execute_tool(
        integration="GITHUB",
        provider="COMPOSIO",
        tool_name="TOOL_1",
        arguments={"arg": "val1"},
    )

    result2 = service.execute_tool(
        integration="SLACK",
        provider="COMPOSIO",
        tool_name="TOOL_2",
        arguments={"arg": "val2"},
    )

    # THEN each tool execution should return its respective result
    assert result1["result"] == "first"
    assert result2["result"] == "second"

    # AND the API should have been called twice
    assert mock_client.integrations.execute_integration_tool.call_count == 2


def test_vellum_integration_service_execute_tool_structured_403_with_integration(vellum_client):
    """Test structured 403 responses with integration field (current backend format)"""
    from vellum.client.core.api_error import ApiError
    from vellum.workflows.errors.types import WorkflowErrorCode

    # GIVEN a mock client configured to raise a structured 403 error with integration
    mock_client = vellum_client
    mock_client.integrations = mock.MagicMock()

    # Mock current backend structure with integration as direct field
    structured_error_body = {
        "code": "INTEGRATION_CREDENTIALS_UNAVAILABLE",
        "message": "You must authenticate with this integration before you can execute this tool.",
        "integration": {
            "id": "550e8400-e29b-41d4-a716-446655440000",
            "provider": "COMPOSIO",
            "name": "GITHUB",
        },
    }
    mock_client.integrations.execute_integration_tool.side_effect = ApiError(
        status_code=403,
        body=structured_error_body,
    )

    # WHEN we attempt to execute a tool without credentials
    service = VellumIntegrationService(client=mock_client)
    with pytest.raises(NodeException) as exc_info:
        service.execute_tool(
            integration="GITHUB",
            provider="COMPOSIO",
            tool_name="GITHUB_CREATE_AN_ISSUE",
            arguments={"repo": "user/repo"},
        )

    # THEN it should raise NodeException with INTEGRATION_CREDENTIALS_UNAVAILABLE code
    assert exc_info.value.code == WorkflowErrorCode.INTEGRATION_CREDENTIALS_UNAVAILABLE

    # AND the error message should match the backend response
    assert "You must authenticate with this integration" in exc_info.value.message

    # AND raw_data should contain integration details nested under "integration" key
    assert exc_info.value.raw_data is not None
    assert exc_info.value.raw_data["integration"]["id"] == "550e8400-e29b-41d4-a716-446655440000"
    assert exc_info.value.raw_data["integration"]["name"] == "GITHUB"
    assert exc_info.value.raw_data["integration"]["provider"] == "COMPOSIO"


def test_vellum_integration_service_execute_tool_legacy_403_error(vellum_client):
    """Test backward compatibility with legacy 403 responses (before PR #14857)"""
    from vellum.client.core.api_error import ApiError
    from vellum.workflows.errors.types import WorkflowErrorCode

    # GIVEN a mock client configured to raise a legacy 403 error
    mock_client = vellum_client
    mock_client.integrations = mock.MagicMock()

    legacy_error_body = {"detail": "You do not have permission to execute this tool."}
    mock_client.integrations.execute_integration_tool.side_effect = ApiError(
        status_code=403,
        body=legacy_error_body,
    )

    # WHEN we attempt to execute a tool that returns a legacy 403 error
    service = VellumIntegrationService(client=mock_client)
    with pytest.raises(NodeException) as exc_info:
        service.execute_tool(
            integration="GITHUB",
            provider="COMPOSIO",
            tool_name="GITHUB_CREATE_AN_ISSUE",
            arguments={"repo": "user/repo"},
        )

    # THEN it should use the generic PROVIDER_CREDENTIALS_UNAVAILABLE code
    assert exc_info.value.code == WorkflowErrorCode.PROVIDER_CREDENTIALS_UNAVAILABLE

    # AND the message should match the legacy format
    assert "You do not have permission" in exc_info.value.message

    # AND raw_data should be None (no integration details available)
    assert exc_info.value.raw_data is None



import pytest

from vellum.workflows.descriptors.exceptions import InvalidExpressionException
from vellum.workflows.expressions.minus import MinusExpression
from vellum.workflows.state.base import BaseState


class TestState(BaseState):
    number_value: int = 10
    float_value: float = 15.5


def test_minus_expression_numbers():
    """
    Tests that MinusExpression correctly subtracts two numbers.
    """

    state = TestState()

    expression = TestState.number_value.minus(3)

    result = expression.resolve(state)
    assert result == 7


def test_minus_expression_floats():
    """
    Tests that MinusExpression correctly subtracts two floats.
    """

    state = TestState()

    expression = TestState.float_value.minus(5.5)

    result = expression.resolve(state)
    assert result == 10.0


def test_minus_expression_unsupported_type():
    """
    Tests that MinusExpression raises an exception for unsupported types.
    """

    class NoSubSupport:
        pass

    no_sub_obj = NoSubSupport()
    expression = MinusExpression(lhs=no_sub_obj, rhs=5)
    state = TestState()

    with pytest.raises(InvalidExpressionException, match="'NoSubSupport' must support the '-' operator"):
        expression.resolve(state)


def test_minus_expression_name():
    """
    Tests that MinusExpression has the correct name.
    """

    expression = MinusExpression(lhs=10, rhs=3)

    assert expression.name == "10 - 3"


def test_minus_expression_types():
    """
    Tests that MinusExpression has the correct types.
    """

    expression = MinusExpression(lhs=10, rhs=3)

    assert expression.types == (object,)


def test_sub_operator_numbers():
    """
    Tests that __sub__ operator correctly subtracts two numbers.
    """

    state = TestState()

    expression = TestState.number_value - 3

    result = expression.resolve(state)
    assert result == 7


def test_sub_operator_floats():
    """
    Tests that __sub__ operator correctly subtracts two floats.
    """

    state = TestState()

    expression = TestState.float_value - 5.5

    result = expression.resolve(state)
    assert result == 10.0


def test_sub_operator_returns_minus_expression():
    """
    Tests that __sub__ operator returns a MinusExpression.
    """

    expression = TestState.number_value - 3

    # THEN it should return a MinusExpression
    assert isinstance(expression, MinusExpression)



# This file was auto-generated by Fern from our API Definition.

import typing
from json.decoder import JSONDecodeError

from ...core.api_error import ApiError
from ...core.client_wrapper import AsyncClientWrapper, SyncClientWrapper
from ...core.http_response import AsyncHttpResponse, HttpResponse
from ...core.jsonable_encoder import jsonable_encoder
from ...core.pydantic_utilities import parse_obj_as
from ...core.request_options import RequestOptions
from ...types.paginated_workflow_sandbox_example_list import PaginatedWorkflowSandboxExampleList
from ...types.workflow_deployment_read import WorkflowDeploymentRead
from .types.list_workflow_sandbox_examples_request_tag import ListWorkflowSandboxExamplesRequestTag

# this is used as the default value for optional parameters
OMIT = typing.cast(typing.Any, ...)


class RawWorkflowSandboxesClient:
    def __init__(self, *, client_wrapper: SyncClientWrapper):
        self._client_wrapper = client_wrapper

    def deploy_workflow(
        self,
        id: str,
        *,
        workflow_deployment_id: typing.Optional[str] = OMIT,
        workflow_deployment_name: typing.Optional[str] = OMIT,
        label: typing.Optional[str] = OMIT,
        release_tags: typing.Optional[typing.Sequence[str]] = OMIT,
        release_description: typing.Optional[str] = OMIT,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> HttpResponse[WorkflowDeploymentRead]:
        """
        Parameters
        ----------
        id : str
            A UUID string identifying this workflow sandbox.

        workflow_deployment_id : typing.Optional[str]
            The Vellum-generated ID of the Workflow Deployment you'd like to update. Cannot specify both this and workflow_deployment_name. Leave null to create a new Workflow Deployment.

        workflow_deployment_name : typing.Optional[str]
            The unique name of the Workflow Deployment you'd like to either create or update. Cannot specify both this and workflow_deployment_id. If provided and matches an existing Workflow Deployment, that Workflow Deployment will be updated. Otherwise, a new Prompt Deployment will be created.

        label : typing.Optional[str]
            In the event that a new Workflow Deployment is created, this will be the label it's given.

        release_tags : typing.Optional[typing.Sequence[str]]
            Optionally provide the release tags that you'd like to be associated with the latest release of the created/updated Prompt Deployment.

        release_description : typing.Optional[str]
            Optionally provide a description that details what's new in this Release.

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        HttpResponse[WorkflowDeploymentRead]

        """
        _response = self._client_wrapper.httpx_client.request(
            f"v1/workflow-sandboxes/{jsonable_encoder(id)}/deploy",
            base_url=self._client_wrapper.get_environment().default,
            method="POST",
            json={
                "workflow_deployment_id": workflow_deployment_id,
                "workflow_deployment_name": workflow_deployment_name,
                "label": label,
                "release_tags": release_tags,
                "release_description": release_description,
            },
            headers={
                "content-type": "application/json",
            },
            request_options=request_options,
            omit=OMIT,
        )
        try:
            if 200 <= _response.status_code < 300:
                _data = typing.cast(
                    WorkflowDeploymentRead,
                    parse_obj_as(
                        type_=WorkflowDeploymentRead,  # type: ignore
                        object_=_response.json(),
                    ),
                )
                return HttpResponse(response=_response, data=_data)
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response.text)
        raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response_json)

    def list_workflow_sandbox_examples(
        self,
        *,
        limit: typing.Optional[int] = None,
        offset: typing.Optional[int] = None,
        ordering: typing.Optional[str] = None,
        tag: typing.Optional[ListWorkflowSandboxExamplesRequestTag] = None,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> HttpResponse[PaginatedWorkflowSandboxExampleList]:
        """
        List Workflow Sandbox examples that were previously cloned into the User's Workspace

        Parameters
        ----------
        limit : typing.Optional[int]
            Number of results to return per page.

        offset : typing.Optional[int]
            The initial index from which to return the results.

        ordering : typing.Optional[str]
            Which field to use when ordering the results.

        tag : typing.Optional[ListWorkflowSandboxExamplesRequestTag]

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        HttpResponse[PaginatedWorkflowSandboxExampleList]

        """
        _response = self._client_wrapper.httpx_client.request(
            "v1/workflow-sandboxes/examples",
            base_url=self._client_wrapper.get_environment().default,
            method="GET",
            params={
                "limit": limit,
                "offset": offset,
                "ordering": ordering,
                "tag": tag,
            },
            request_options=request_options,
        )
        try:
            if 200 <= _response.status_code < 300:
                _data = typing.cast(
                    PaginatedWorkflowSandboxExampleList,
                    parse_obj_as(
                        type_=PaginatedWorkflowSandboxExampleList,  # type: ignore
                        object_=_response.json(),
                    ),
                )
                return HttpResponse(response=_response, data=_data)
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response.text)
        raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response_json)


class AsyncRawWorkflowSandboxesClient:
    def __init__(self, *, client_wrapper: AsyncClientWrapper):
        self._client_wrapper = client_wrapper

    async def deploy_workflow(
        self,
        id: str,
        *,
        workflow_deployment_id: typing.Optional[str] = OMIT,
        workflow_deployment_name: typing.Optional[str] = OMIT,
        label: typing.Optional[str] = OMIT,
        release_tags: typing.Optional[typing.Sequence[str]] = OMIT,
        release_description: typing.Optional[str] = OMIT,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> AsyncHttpResponse[WorkflowDeploymentRead]:
        """
        Parameters
        ----------
        id : str
            A UUID string identifying this workflow sandbox.

        workflow_deployment_id : typing.Optional[str]
            The Vellum-generated ID of the Workflow Deployment you'd like to update. Cannot specify both this and workflow_deployment_name. Leave null to create a new Workflow Deployment.

        workflow_deployment_name : typing.Optional[str]
            The unique name of the Workflow Deployment you'd like to either create or update. Cannot specify both this and workflow_deployment_id. If provided and matches an existing Workflow Deployment, that Workflow Deployment will be updated. Otherwise, a new Prompt Deployment will be created.

        label : typing.Optional[str]
            In the event that a new Workflow Deployment is created, this will be the label it's given.

        release_tags : typing.Optional[typing.Sequence[str]]
            Optionally provide the release tags that you'd like to be associated with the latest release of the created/updated Prompt Deployment.

        release_description : typing.Optional[str]
            Optionally provide a description that details what's new in this Release.

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        AsyncHttpResponse[WorkflowDeploymentRead]

        """
        _response = await self._client_wrapper.httpx_client.request(
            f"v1/workflow-sandboxes/{jsonable_encoder(id)}/deploy",
            base_url=self._client_wrapper.get_environment().default,
            method="POST",
            json={
                "workflow_deployment_id": workflow_deployment_id,
                "workflow_deployment_name": workflow_deployment_name,
                "label": label,
                "release_tags": release_tags,
                "release_description": release_description,
            },
            headers={
                "content-type": "application/json",
            },
            request_options=request_options,
            omit=OMIT,
        )
        try:
            if 200 <= _response.status_code < 300:
                _data = typing.cast(
                    WorkflowDeploymentRead,
                    parse_obj_as(
                        type_=WorkflowDeploymentRead,  # type: ignore
                        object_=_response.json(),
                    ),
                )
                return AsyncHttpResponse(response=_response, data=_data)
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response.text)
        raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response_json)

    async def list_workflow_sandbox_examples(
        self,
        *,
        limit: typing.Optional[int] = None,
        offset: typing.Optional[int] = None,
        ordering: typing.Optional[str] = None,
        tag: typing.Optional[ListWorkflowSandboxExamplesRequestTag] = None,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> AsyncHttpResponse[PaginatedWorkflowSandboxExampleList]:
        """
        List Workflow Sandbox examples that were previously cloned into the User's Workspace

        Parameters
        ----------
        limit : typing.Optional[int]
            Number of results to return per page.

        offset : typing.Optional[int]
            The initial index from which to return the results.

        ordering : typing.Optional[str]
            Which field to use when ordering the results.

        tag : typing.Optional[ListWorkflowSandboxExamplesRequestTag]

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        AsyncHttpResponse[PaginatedWorkflowSandboxExampleList]

        """
        _response = await self._client_wrapper.httpx_client.request(
            "v1/workflow-sandboxes/examples",
            base_url=self._client_wrapper.get_environment().default,
            method="GET",
            params={
                "limit": limit,
                "offset": offset,
                "ordering": ordering,
                "tag": tag,
            },
            request_options=request_options,
        )
        try:
            if 200 <= _response.status_code < 300:
                _data = typing.cast(
                    PaginatedWorkflowSandboxExampleList,
                    parse_obj_as(
                        type_=PaginatedWorkflowSandboxExampleList,  # type: ignore
                        object_=_response.json(),
                    ),
                )
                return AsyncHttpResponse(response=_response, data=_data)
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response.text)
        raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response_json)



from deepdiff import DeepDiff

from vellum_ee.workflows.display.workflows.get_vellum_workflow_display_class import get_workflow_display

from tests.workflows.basic_guardrail_node.workflow import BasicGuardrailNodeWorkflow


def test_serialize_workflow():
    # GIVEN a workflow that uses a guardrail node
    # WHEN we serialize it
    workflow_display = get_workflow_display(workflow_class=BasicGuardrailNodeWorkflow)

    serialized_workflow: dict = workflow_display.serialize()

    # THEN we should get a serialized representation of the workflow
    assert serialized_workflow.keys() == {
        "workflow_raw_data",
        "input_variables",
        "state_variables",
        "output_variables",
    }

    # AND its input variables should be what we expect
    input_variables = serialized_workflow["input_variables"]
    assert len(input_variables) == 2
    assert not DeepDiff(
        [
            {
                "id": "eb1b1913-9fb8-4b8c-8901-09d9b9edc1c3",
                "key": "actual",
                "type": "STRING",
                "required": True,
                "default": None,
                "extensions": {"color": None},
            },
            {
                "id": "545ff95e-e86f-4d06-a991-602781e72605",
                "key": "expected",
                "type": "STRING",
                "required": True,
                "default": None,
                "extensions": {"color": None},
            },
        ],
        input_variables,
        ignore_order=True,
    )

    # AND its output variables should be what we expect
    output_variables = serialized_workflow["output_variables"]
    assert len(output_variables) == 1
    assert output_variables == [{"id": "2abd2b3b-c301-4834-a43f-5db3604f8422", "key": "score", "type": "NUMBER"}]

    # AND its raw data is what we expect
    workflow_raw_data = serialized_workflow["workflow_raw_data"]
    assert len(workflow_raw_data["nodes"]) == 3
    assert len(workflow_raw_data["edges"]) == 2

    # AND each node should be serialized correctly
    entrypoint_node = workflow_raw_data["nodes"][0]
    assert entrypoint_node == {
        "id": "54c5c7d0-ab86-4ae9-b0b8-ea9ca7b87c14",
        "type": "ENTRYPOINT",
        "base": None,
        "definition": None,
        "inputs": [],
        "data": {"label": "Entrypoint Node", "source_handle_id": "41840690-8d85-486e-a864-b0661ccf0f2e"},
        "display_data": {"position": {"x": 0.0, "y": -50.0}},
    }

    guardrail_node = workflow_raw_data["nodes"][1]
    assert guardrail_node == {
        "id": "7fef2bbc-cdfc-4f66-80eb-2a52ee52da5f",
        "type": "METRIC",
        "inputs": [
            {
                "id": "dad627f3-d46d-4f12-b3d0-4aa25a5f24b5",
                "key": "expected",
                "value": {
                    "rules": [
                        {
                            "type": "INPUT_VARIABLE",
                            "data": {"input_variable_id": "545ff95e-e86f-4d06-a991-602781e72605"},
                        }
                    ],
                    "combinator": "OR",
                },
            },
            {
                "id": "42aef2a5-5dcf-41ea-8da4-1eee1f8baf84",
                "key": "actual",
                "value": {
                    "rules": [
                        {
                            "type": "INPUT_VARIABLE",
                            "data": {"input_variable_id": "eb1b1913-9fb8-4b8c-8901-09d9b9edc1c3"},
                        }
                    ],
                    "combinator": "OR",
                },
            },
        ],
        "data": {
            "label": "Example Guardrail Node",
            "source_handle_id": "baa8baa7-8849-4b96-a90d-c0545a60d3a8",
            "target_handle_id": "53c299c7-1df2-4d54-bb0d-559a4947c16d",
            "error_output_id": None,
            "metric_definition_id": "example_metric_definition",
            "release_tag": "LATEST",
        },
        "display_data": {"position": {"x": 200.0, "y": -50.0}},
        "base": {
            "module": ["vellum", "workflows", "nodes", "displayable", "guardrail_node", "node"],
            "name": "GuardrailNode",
        },
        "definition": {
            "module": ["tests", "workflows", "basic_guardrail_node", "workflow"],
            "name": "ExampleGuardrailNode",
        },
        "trigger": {
            "id": "53c299c7-1df2-4d54-bb0d-559a4947c16d",
            "merge_behavior": "AWAIT_ANY",
        },
        "ports": [{"id": "baa8baa7-8849-4b96-a90d-c0545a60d3a8", "name": "default", "type": "DEFAULT"}],
    }

    final_output_node = workflow_raw_data["nodes"][2]
    assert final_output_node == {
        "id": "cbc7197e-67c9-4af5-b781-879c8fd3e4c9",
        "type": "TERMINAL",
        "base": {
            "module": [
                "vellum",
                "workflows",
                "nodes",
                "displayable",
                "final_output_node",
                "node",
            ],
            "name": "FinalOutputNode",
        },
        "definition": None,
        "data": {
            "label": "Final Output",
            "name": "score",
            "target_handle_id": "001b97f6-2bc8-4d1e-9572-028dcf17df4e",
            "output_id": "2abd2b3b-c301-4834-a43f-5db3604f8422",
            "output_type": "NUMBER",
            "node_input_id": "6321442a-0d0d-4e25-965d-c24ff24712c5",
        },
        "inputs": [
            {
                "id": "6321442a-0d0d-4e25-965d-c24ff24712c5",
                "key": "node_input",
                "value": {
                    "rules": [
                        {
                            "type": "NODE_OUTPUT",
                            "data": {
                                "node_id": "7fef2bbc-cdfc-4f66-80eb-2a52ee52da5f",
                                "output_id": "eb9c8043-4c04-467b-944e-633c54de6876",
                            },
                        }
                    ],
                    "combinator": "OR",
                },
            }
        ],
        "display_data": {"position": {"x": 400.0, "y": -50.0}},
    }

    # AND each edge should be serialized correctly
    serialized_edges = workflow_raw_data["edges"]
    assert serialized_edges == [
        {
            "id": "4737867f-1967-45a1-966b-a0bda81a583d",
            "source_node_id": "54c5c7d0-ab86-4ae9-b0b8-ea9ca7b87c14",
            "source_handle_id": "41840690-8d85-486e-a864-b0661ccf0f2e",
            "target_node_id": "7fef2bbc-cdfc-4f66-80eb-2a52ee52da5f",
            "target_handle_id": "53c299c7-1df2-4d54-bb0d-559a4947c16d",
            "type": "DEFAULT",
        },
        {
            "id": "5c456a17-a92b-4dad-9569-306043707c9f",
            "source_node_id": "7fef2bbc-cdfc-4f66-80eb-2a52ee52da5f",
            "source_handle_id": "baa8baa7-8849-4b96-a90d-c0545a60d3a8",
            "target_node_id": "cbc7197e-67c9-4af5-b781-879c8fd3e4c9",
            "target_handle_id": "001b97f6-2bc8-4d1e-9572-028dcf17df4e",
            "type": "DEFAULT",
        },
    ]

    # AND the display data is what we expect
    display_data = workflow_raw_data["display_data"]
    assert display_data == {"viewport": {"x": 0.0, "y": 0.0, "zoom": 1.0}}

    # AND the definition is what we expect
    definition = workflow_raw_data["definition"]
    assert definition == {
        "name": "BasicGuardrailNodeWorkflow",
        "module": [
            "tests",
            "workflows",
            "basic_guardrail_node",
            "workflow",
        ],
    }
